import time
from http.client import HTTPConnection
from subprocess import PIPE, Popen

import pytest
from playwright.sync_api import Playwright


@pytest.fixture
def dev_server():
    """
    Running dev server and Playwright test in parallel.
    More info https://til.simonwillison.net/pytest/playwright-pytest
    """
    process = Popen(
        ["python", "-m", "piccolo_admin.example"],
        stdout=PIPE,
    )
    retries = 5
    while retries > 0:
        conn = HTTPConnection("localhost:8000")
        try:
            conn.request("HEAD", "/")
            response = conn.getresponse()
            if response is not None:
                yield process
                break
        except ConnectionRefusedError:
            time.sleep(1)
            retries -= 1

    if not retries:
        raise RuntimeError("Failed to start http server")
    else:
        process.terminate()
        process.wait()


def test_login_logout(playwright: Playwright, dev_server) -> None:
    browser = playwright.chromium.launch()
    context = browser.new_context(record_video_dir="videos/")
    # Open new page
    page = context.new_page()
    # Go to http://localhost:8000/#/login?nextURL=%2F
    page.goto("http://localhost:8000/#/login?nextURL=%2F")
    # Click input[name="username"]
    page.locator('input[name="username"]').click()
    # Fill input[name="username"]
    page.locator('input[name="username"]').fill("piccolo")
    # Press Tab
    page.locator('input[name="username"]').press("Tab")
    # Fill input[name="password"]
    page.locator('input[name="password"]').fill("piccolo123")
    # Press Enter
    page.locator('input[name="password"]').press("Enter")
    page.wait_for_url("http://localhost:8000/#/")
    # Click #user
    page.locator("#user").click()
    # Click text=Log out
    page.once("dialog", lambda dialog: dialog.dismiss())
    page.locator("text=Log out").click()
    # ---------------------
    context.close()
    browser.close()


def test_row_listing_filter(playwright: Playwright, dev_server) -> None:
    browser = playwright.chromium.launch()
    context = browser.new_context(record_video_dir="videos/")
    # Open new page
    page = context.new_page()
    # Go to http://localhost:8000/#/login?nextURL=%2F
    page.goto("http://localhost:8000/#/login?nextURL=%2F")
    # Click input[name="username"]
    page.locator('input[name="username"]').click()
    # Fill input[name="username"]
    page.locator('input[name="username"]').fill("piccolo")
    # Press Tab
    page.locator('input[name="username"]').press("Tab")
    # Fill input[name="password"]
    page.locator('input[name="password"]').fill("piccolo123")
    # Press Enter
    page.locator('input[name="password"]').press("Enter")
    page.wait_for_url("http://localhost:8000/#/")
    # Click text=director
    page.locator("text=director").click()
    page.wait_for_url("http://localhost:8000/#/director")
    # Click text=Show filters
    page.locator("text=Show filters").click()
    # Click input[name="name"]
    page.locator('input[name="name"]').click()
    # Fill input[name="name"]
    page.locator('input[name="name"]').fill("Howard")
    # Press Enter
    page.locator('input[name="name"]').press("Enter")
    # Click text=Clear filters
    page.locator("text=Clear filters").click()
    # Click input[name="name"]
    page.locator('input[name="name"]').click()
    # Fill input[name="name"]
    page.locator('input[name="name"]').fill("Ron")
    # Click text=Apply
    page.locator("text=Apply").click()
    # Click text=Clear filters
    page.locator("text=Clear filters").click()
    # Click text=Close
    page.locator("text=Close").click()
    # Click span:has-text("Sort")
    page.locator("span", has_text="Sort").click()
    # Select ascending
    page.locator('select[name="ordering"]').select_option("ascending")
    # Click button:has-text("Sort")
    page.locator("button", has_text="Sort").click()
    page.wait_for_timeout(2000)
    page.locator("#user").click()
    # Click text=Log out
    page.once("dialog", lambda dialog: dialog.accept())
    page.locator("text=Log out").click()
    page.wait_for_url("http://localhost:8000/#/login?nextURL=%2F")
    # ---------------------
    context.close()
    browser.close()


def test_custom_form(playwright: Playwright, dev_server) -> None:
    browser = playwright.chromium.launch()
    context = browser.new_context(record_video_dir="videos/")
    # Open new page
    page = context.new_page()
    # Go to http://localhost:8000/#/login?nextURL=%2F
    page.goto("http://localhost:8000/#/login?nextURL=%2F")
    # Click input[name="username"]
    page.locator('input[name="username"]').click()
    # Fill input[name="username"]
    page.locator('input[name="username"]').fill("piccolo")
    # Press Tab
    page.locator('input[name="username"]').press("Tab")
    # Fill input[name="password"]
    page.locator('input[name="password"]').fill("piccolo123")
    # Press Enter
    page.locator('input[name="password"]').press("Enter")
    page.wait_for_url("http://localhost:8000/#/")
    # Click text=Business email form
    page.locator("text=Business email form").click()
    page.wait_for_url("http://localhost:8000/#/forms/business-email-form")
    # Click input[name="email"]
    page.locator('input[name="email"]').click()
    # Fill input[name="email"]
    page.locator('input[name="email"]').fill("director@director.com")
    # Click input[name="title"]
    page.locator('input[name="title"]').click()
    # Click text=Piccolo Admin piccolo Back Business email form Send an email to a business assoc
    page.locator(
        "text=Piccolo Admin piccolo Back Business email form Send an email to a business assoc"
    ).click()
    # Click input[name="title"]
    page.locator('input[name="title"]').click()
    # Fill input[name="title"]
    page.locator('input[name="title"]').fill("Hello")
    # Click input[name="content"]
    page.locator('input[name="content"]').click()
    # Fill input[name="content"]
    page.locator('input[name="content"]').fill("Hello from Piccolo Admin")
    # Click button:has-text("Submit")
    page.locator('button:has-text("Submit")').click()
    # Click text=Back >> nth=0
    page.locator("text=Back").first.click()
    page.wait_for_url("http://localhost:8000/#/")
    # Click text=Booking form
    page.locator("text=Booking form").click()
    page.wait_for_url("http://localhost:8000/#/forms/booking-form")
    # Click input[name="email"]
    page.locator('input[name="email"]').click()
    # Fill input[name="email"]
    page.locator('input[name="email"]').fill("customer@customer.com")
    # Click input[name="name"]
    page.locator('input[name="name"]').click()
    # Fill input[name="name"]
    page.locator('input[name="name"]').fill("Bob")
    # Click input[name="notes"]
    page.locator('input[name="notes"]').click()
    # Fill input[name="notes"]
    page.locator('input[name="notes"]').fill("Star Wars please")
    # Click button:has-text("Submit")
    page.locator('button:has-text("Submit")').click()
    # Click text=Back >> nth=0
    page.locator("text=Back").first.click()
    page.wait_for_timeout(2000)
    page.locator("#user").click()
    # Click text=Log out
    page.once("dialog", lambda dialog: dialog.accept())
    page.locator("text=Log out").click()
    page.wait_for_url("http://localhost:8000/#/login?nextURL=%2F")
    # ---------------------
    context.close()
    browser.close()


def test_file_upload(playwright: Playwright, dev_server) -> None:
    browser = playwright.chromium.launch()
    context = browser.new_context(record_video_dir="videos/")
    # Open new page
    page = context.new_page()
    # Go to http://localhost:8000/#/login?nextURL=%2F
    page.goto("http://localhost:8000/#/login?nextURL=%2F")
    # Click input[name="username"]
    page.locator('input[name="username"]').click()
    # Fill input[name="username"]
    page.locator('input[name="username"]').fill("piccolo")
    # Press Tab
    page.locator('input[name="username"]').press("Tab")
    # Fill input[name="password"]
    page.locator('input[name="password"]').fill("piccolo123")
    # Press Enter
    page.locator('input[name="password"]').press("Enter")
    page.wait_for_url("http://localhost:8000/#/")
    # Click text=director
    page.locator("text=director").click()
    page.wait_for_url("http://localhost:8000/#/director")
    # Click a:has-text("8")
    page.locator('a:has-text("8")').click()
    page.wait_for_url("http://localhost:8000/#/director/8")
    # Upload piccolo.jpg
    page.locator('input[type="file"]').set_input_files(
        "./e2e/upload/piccolo.jpg"
    )
    # Click text=Save
    page.locator("text=Save").click()
    # Click text=Back
    page.locator("text=Back").click()
    page.wait_for_url("http://localhost:8000/#/director")
    # Click image link
    page.locator("text=piccolo-").click()
    # Close image preview
    page.locator("div#media_viewer div.top_bar p.close").click()
    # ---------------------
    context.close()
    browser.close()


def test_bulk_update(playwright: Playwright, dev_server) -> None:
    browser = playwright.chromium.launch()
    context = browser.new_context(record_video_dir="videos/")
    # Open new page
    page = context.new_page()
    # Go to http://localhost:8000/#/login?nextURL=%2F
    page.goto("http://localhost:8000/#/login?nextURL=%2F")
    # Click input[name="username"]
    page.locator('input[name="username"]').click()
    # Fill input[name="username"]
    page.locator('input[name="username"]').fill("piccolo")
    # Press Tab
    page.locator('input[name="username"]').press("Tab")
    # Fill input[name="password"]
    page.locator('input[name="password"]').fill("piccolo123")
    # Press Enter
    page.locator('input[name="password"]').press("Enter")
    page.wait_for_url("http://localhost:8000/#/")
    # Click text=director
    page.locator("text=director").click()
    page.wait_for_url("http://localhost:8000/#/director")
    # Check text=id Name Gender Photo >> input[type="checkbox"]
    page.locator('text=id Name Gender Photo >> input[type="checkbox"]').check()
    # Click a:has-text("Update 8 rows")
    page.locator('a:has-text("Update 8 rows")').click()
    # Select gender
    page.locator('select[name="property"]').select_option("gender")
    # Select f
    page.locator('select[name="gender"]').select_option("f")
    # Click button:has-text("Update")
    page.locator('button:has-text("Update")').click()
    # Click #user
    page.wait_for_timeout(2000)
    page.locator("#user").click()
    # Click text=Log out
    page.once("dialog", lambda dialog: dialog.accept())
    page.locator("text=Log out").click()
    page.wait_for_url("http://localhost:8000/#/login?nextURL=%2F")
    # ---------------------
    context.close()
    browser.close()


def test_table_crud(playwright: Playwright, dev_server) -> None:
    browser = playwright.chromium.launch()
    context = browser.new_context(record_video_dir="videos/")
    # Open new page
    page = context.new_page()
    # Go to http://localhost:8000/#/login?nextURL=%2F
    page.goto("http://localhost:8000/#/login?nextURL=%2F")
    # Click input[name="username"]
    page.locator('input[name="username"]').click()
    # Fill input[name="username"]
    page.locator('input[name="username"]').fill("piccolo")
    # Press Tab
    page.locator('input[name="username"]').press("Tab")
    # Fill input[name="password"]
    page.locator('input[name="password"]').fill("piccolo123")
    # Press Enter
    page.locator('input[name="password"]').press("Enter")
    page.wait_for_url("http://localhost:8000/#/")
    # Click text=director
    page.locator("text=director").click()
    page.wait_for_url("http://localhost:8000/#/director")
    # Click a:has-text("8")
    page.locator('a:has-text("8")').click()
    page.wait_for_url("http://localhost:8000/#/director/8")
    # Click text=Piccolo Admin piccolo Back Edit director Name Years Nominated Which years this d
    page.locator(
        "text=Piccolo Admin piccolo Back Edit director Name Years Nominated Which years this d"
    ).click()
    # Fill input[type="text"] >> nth=2
    page.locator('input[type="text"]').nth(2).fill("Ronald William Howard2009")
    # Press z with modifiers
    page.locator('input[type="text"]').nth(2).press("Control+z")
    # Click input[name="name"]
    page.locator('input[name="name"]').click()
    # Fill input[name="name"]
    page.locator('input[name="name"]').fill("Ronald William Howard")
    # Click text=Save
    page.locator("text=Save").click()
    # Click text=Back
    page.locator("text=Back").click()
    page.wait_for_url("http://localhost:8000/#/director")
    # Click text=Add Row
    page.locator("text=Add Row").click()
    page.wait_for_url("http://localhost:8000/#/director/add")
    # Click input[name="name"]
    page.locator('input[name="name"]').click()
    # Fill input[name="name"]
    page.locator('input[name="name"]').fill("Emerald Fennell")
    # Click text=AddAdd
    page.get_by_role("link", name="Add Add").first.click()
    page.get_by_role("listitem").filter(has_text="Remove").get_by_role(
        "textbox"
    ).click()
    page.get_by_role("listitem").filter(has_text="Remove").get_by_role(
        "textbox"
    ).fill("2022")
    page.get_by_role("link", name="Remove").click()
    # Select f
    page.locator('select[name="gender"]').select_option("f")
    # Click text=Create
    page.locator("text=Create").click()
    # Click text=Back
    page.locator("text=Back").click()
    page.wait_for_timeout(3000)
    page.wait_for_url("http://localhost:8000/#/director")
    # Check text=5Irvin Kershner Male >> input[type="checkbox"]
    page.locator('text=5Irvin Kershner Male >> input[type="checkbox"]').check()
    # Click text=Delete 1 rows
    page.once("dialog", lambda dialog: dialog.accept())
    page.locator("text=Delete 1 rows").click()
    page.wait_for_timeout(3000)
    # Check text=id Name Gender Photo >> input[type="checkbox"]
    page.locator('text=id Name Gender Photo >> input[type="checkbox"]').check()
    # Click text=Delete 8 rows
    page.once("dialog", lambda dialog: dialog.accept())
    page.locator("text=Delete 8 rows").click()
    # wait for rows deletion
    page.wait_for_timeout(2000)
    page.locator("#user").click()
    # Click text=Log out
    page.once("dialog", lambda dialog: dialog.dismiss())
    page.locator("text=Log out").click()
    # ---------------------
    context.close()
    browser.close()



"""
Creates a basic wrapper around a Piccolo model, turning it into an ASGI app.
"""
from __future__ import annotations

import inspect
import itertools
import json
import logging
import os
import typing as t
from dataclasses import dataclass
from datetime import timedelta
from functools import partial

from fastapi import FastAPI, File, Form, UploadFile
from piccolo.apps.user.tables import BaseUser
from piccolo.columns.base import Column
from piccolo.columns.reference import LazyTableReference
from piccolo.table import Table
from piccolo.utils.warnings import Level, colored_warning
from piccolo_api.change_password.endpoints import change_password
from piccolo_api.crud.endpoints import PiccoloCRUD
from piccolo_api.crud.hooks import Hook
from piccolo_api.crud.validators import Validators
from piccolo_api.csrf.middleware import CSRFMiddleware
from piccolo_api.fastapi.endpoints import FastAPIKwargs, FastAPIWrapper
from piccolo_api.media.base import MediaStorage
from piccolo_api.media.local import LocalMediaStorage
from piccolo_api.openapi.endpoints import swagger_ui
from piccolo_api.rate_limiting.middleware import (
    InMemoryLimitProvider,
    RateLimitingMiddleware,
    RateLimitProvider,
)
from piccolo_api.session_auth.endpoints import session_login, session_logout
from piccolo_api.session_auth.middleware import SessionsAuthBackend
from piccolo_api.session_auth.tables import SessionsBase
from pydantic import BaseModel, Field, ValidationError
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse
from starlette.staticfiles import StaticFiles

from .translations.data import TRANSLATIONS
from .translations.models import (
    Translation,
    TranslationListItem,
    TranslationListResponse,
)
from .version import __VERSION__ as PICCOLO_ADMIN_VERSION

logger = logging.getLogger(__name__)


ASSET_PATH = os.path.join(os.path.dirname(__file__), "dist")


class UserResponseModel(BaseModel):
    username: str
    user_id: str


class MetaResponseModel(BaseModel):
    piccolo_admin_version: str
    site_name: str


class StoreFileResponseModel(BaseModel):
    file_key: str = Field(description="For example `my_file-some-uuid.jpeg`.")


class GenerateFileURLRequestModel(BaseModel):
    column_name: str
    table_name: str
    file_key: str = Field(description="For example `my_file-some-uuid.jpeg`.")


class GenerateFileURLResponseModel(BaseModel):
    file_url: str = Field(description="A URL which the file is accessible on.")


@dataclass
class TableConfig:
    """
    Gives the user more control over how a ``Table`` appears in the UI.

    :param table_class:
        The :class:`Table <piccolo.table.Table>` class to configure.
    :param visible_columns:
        If specified, only these columns will be shown in the list view of the
        UI. This is useful when you have a lot of columns.
    :param exclude_visible_columns:
        You can specify this instead of ``visible_columns``, in which case all
        of the ``Table`` columns except the ones specified will be shown in the
        list view.
    :param visible_filters:
        If specified, only these columns will be shown in the filter sidebar
        of the UI. This is useful when you have a lot of columns.
    :param exclude_visible_filters:
        You can specify this instead of ``visible_filters``, in which case all
        of the ``Table`` columns except the ones specified will be shown in the
        filter sidebar.
    :param rich_text_columns:
        You can specify ``rich_text_columns`` if you want a WYSIWYG editor
        on certain Piccolo :class:`Text <piccolo.columns.column_types.Text>`
        columns. Any columns not specified will use a standard HTML textarea
        tag in the UI.
    :param hooks:
        These are passed directly to
        :class:`PiccoloCRUD <piccolo_api.crud.endpoints.PiccoloCRUD>`, which
        powers Piccolo Admin under the hood. It allows you to run custom logic
        when a row is modified.
    :param media_storage:
        These columns will be used to store media. We don't directly store the
        media in the database, but instead store a string, which is a unique
        identifier, and can be used to retrieve a URL for accessing the file.
        Piccolo Admin automatically renders a file upload widget for each media
        column in the UI.
    :param validators:
        The :class:`Validators <piccolo_api.crud.endpoints.Validators>` are
        passed directly to
        :class:`PiccoloCRUD <piccolo_api.crud.endpoints.PiccoloCRUD>`, which
        powers Piccolo Admin under the hood. It allows fine grained access
        control over each API endpoint. For example, limiting which users can
        ``POST`` data::

            from piccolo_api.crud.endpoints import PiccoloCRUD
            from starlette.exceptions import HTTPException
            from starlette.requests import Request


            async def manager_only(
                piccolo_crud: PiccoloCRUD,
                request: Request
            ):
                # The Piccolo `BaseUser` can be accessed from the request.
                user = request.user.user

                # Assuming we have another database table where we record
                # users with certain permissions.
                manager = await Manager.exists().where(manager.user == user)

                if not manager:
                    # Raise a Starlette exception if we want to reject the
                    # request.
                    raise HTTPException(
                        status_code=403,
                        detail="Only managers are allowed to do this"
                    )

            admin = create_admin(
                tables=TableConfig(
                    Movie,
                    validators=Validators(post_single=[manager_only])
                )
            )

    """

    table_class: t.Type[Table]
    visible_columns: t.Optional[t.List[Column]] = None
    exclude_visible_columns: t.Optional[t.List[Column]] = None
    visible_filters: t.Optional[t.List[Column]] = None
    exclude_visible_filters: t.Optional[t.List[Column]] = None
    rich_text_columns: t.Optional[t.List[Column]] = None
    hooks: t.Optional[t.List[Hook]] = None
    media_storage: t.Optional[t.Sequence[MediaStorage]] = None
    validators: t.Optional[Validators] = None

    def __post_init__(self):
        if self.visible_columns and self.exclude_visible_columns:
            raise ValueError(
                "Only specify `visible_columns` or "
                "`exclude_visible_columns`."
            )

        if self.visible_filters and self.exclude_visible_filters:
            raise ValueError(
                "Only specify `visible_filters` or `exclude_visible_filters`."
            )

        # Create a mapping for faster lookups
        self.media_columns = (
            {i.column: i for i in self.media_storage}
            if self.media_storage
            else None
        )

    def _get_columns(
        self,
        include_columns: t.Optional[t.List[Column]],
        exclude_columns: t.Optional[t.List[Column]],
        all_columns: t.List[Column],
    ) -> t.List[Column]:
        if include_columns and not exclude_columns:
            return include_columns

        if exclude_columns and not include_columns:
            column_names = [i._meta.name for i in exclude_columns]
            return [i for i in all_columns if i._meta.name not in column_names]

        return all_columns

    def get_visible_columns(self) -> t.List[Column]:
        return self._get_columns(
            include_columns=self.visible_columns,
            exclude_columns=self.exclude_visible_columns,
            all_columns=self.table_class._meta.columns,
        )

    def get_visible_column_names(self) -> t.Tuple[str, ...]:
        return tuple(i._meta.name for i in self.get_visible_columns())

    def get_visible_filters(self) -> t.List[Column]:
        return self._get_columns(
            include_columns=self.visible_filters,
            exclude_columns=self.exclude_visible_filters,
            all_columns=self.table_class._meta.columns,
        )

    def get_visible_filter_names(self) -> t.Tuple[str, ...]:
        return tuple(i._meta.name for i in self.get_visible_filters())

    def get_rich_text_columns_names(self) -> t.Tuple[str, ...]:
        return (
            tuple(i._meta.name for i in self.rich_text_columns)
            if self.rich_text_columns
            else ()
        )

    def get_media_columns_names(self) -> t.Tuple[str, ...]:
        return (
            tuple(i._meta.name for i in self.media_columns)
            if self.media_columns
            else ()
        )


@dataclass
class FormConfig:
    """
    Used to specify forms, which are passed into ``create_admin``.

    :param name:
        This will be displayed in the UI in the sidebar.
    :param pydantic_model:
        This determines which fields to display in the form, and is used to
        deserialise the responses.
    :param endpoint:
        Your custom handler, which accepts two arguments - the FastAPI /
        Starlette request object, in case you want to access the cookies /
        headers / logged in user (via `request.user`). And secondly an instance
        of the Pydantic model. If it returns a string, it will be shown to
        the user in the UI as the success message. For example ``'Successfully
        sent email'``. The endpoint can be a normal function or async function.
    :param description:
        An optional description which is shown in the UI to explain to the user
        what the form is for.

    Here's a full example:

    .. code-block:: python

        class MyModel(pydantic.BaseModel):
            message: str = "hello world"


        def my_endpoint(request: Request, data: MyModel):
            print(f"I received {data.message}")

            # If we're not happy with the data raise a ValueError
            # The message inside the exception will be displayed in the UI.
            raise ValueError("We were unable to process the form.")

            # If we're happy with the data, just return a string, which
            # will be displayed in the UI.
            return "Successful."


        config = FormConfig(
            name="My Form",
            pydantic_model=MyModel,
            endpoint=my_endpoint
        )

    """

    name: str
    pydantic_model: t.Type[BaseModel]
    endpoint: t.Callable[
        [Request, BaseModel],
        t.Union[str, None, t.Coroutine],
    ]
    description: t.Optional[str] = None

    def __post_init__(self):
        self.slug = self.name.replace(" ", "-").lower()


class FormConfigResponseModel(BaseModel):
    name: str
    slug: str
    description: t.Optional[str] = None


def handle_auth_exception(request: Request, exc: Exception):
    return JSONResponse({"error": "Auth failed"}, status_code=401)


def superuser_validators(piccolo_crud: PiccoloCRUD, request: Request):
    """
    We need to provide extra validation on certain tables (e.g. user and
    sessions), so only superusers can perform certain actions, otherwise the
    security of the application can be compromised.
    """
    user: BaseUser = request.user.user
    if user.superuser:
        return
    if request.method.upper() in ["PUT", "PATCH", "DELETE", "POST"]:
        raise HTTPException(
            status_code=405,
            detail="Only superusers can perform these actions.",
        )


async def log_error(request: Request, exc: HTTPException):
    logger.exception("Server error")
    raise exc


class AdminRouter(FastAPI):
    """
    The root returns a single page app. The other URLs are REST endpoints.
    """

    table: t.List[Table] = []
    auth_table: t.Type[BaseUser] = BaseUser
    template: str = ""

    def __init__(
        self,
        *tables: t.Union[t.Type[Table], TableConfig],
        forms: t.List[FormConfig] = [],
        auth_table: t.Type[BaseUser] = BaseUser,
        session_table: t.Type[SessionsBase] = SessionsBase,
        session_expiry: timedelta = timedelta(hours=1),
        max_session_expiry: timedelta = timedelta(days=7),
        increase_expiry: t.Optional[timedelta] = timedelta(minutes=20),
        page_size: int = 15,
        read_only: bool = False,
        rate_limit_provider: t.Optional[RateLimitProvider] = None,
        production: bool = False,
        site_name: str = "Piccolo Admin",
        default_language_code: str = "auto",
        translations: t.List[Translation] = None,
        allowed_hosts: t.Sequence[str] = [],
        debug: bool = False,
    ) -> None:
        super().__init__(
            title=site_name,
            description=f"{site_name} documentation",
            middleware=[
                Middleware(CSRFMiddleware, allowed_hosts=allowed_hosts)
            ],
            debug=debug,
            exception_handlers={500: log_error},
            docs_url=None,
            redoc_url=None,
        )

        #######################################################################
        # Convert any table arguments which are plain ``Table`` classes into
        # ``TableConfig`` instances.

        table_configs: t.List[TableConfig] = []

        for table in tables:
            if isinstance(table, TableConfig):
                table_configs.append(table)
            else:
                table_configs.append(TableConfig(table_class=table))

        self.table_configs = table_configs
        self.table_config_map = {
            table_config.table_class._meta.tablename: table_config
            for table_config in table_configs
        }

        #######################################################################
        # Make sure columns are configured properly.

        for table_config in table_configs:
            table_class = table_config.table_class
            for column in table_class._meta.columns:
                if column._meta.secret and column._meta.required:
                    message = (
                        f"{table_class._meta.tablename}."
                        f"{column._meta._name} is using `secret` and "
                        f"`required` column args which are incompatible. "
                        f"You may encounter unexpected behavior when using "
                        f"this table within Piccolo Admin."
                    )
                    colored_warning(message, level=Level.high)

        #######################################################################
        # Make sure media storage is configured properly.

        media_storage = [
            i
            for i in itertools.chain(
                *[
                    table_config.media_storage or []
                    for table_config in table_configs
                ]
            )
        ]

        if len(media_storage) != len(set(media_storage)):
            raise ValueError(
                "Media storage is misconfigured - multiple columns are saving "
                "to the same location."
            )

        #######################################################################

        self.default_language_code = default_language_code
        self.translations_map = {
            translation.language_code.lower(): translation
            for translation in (translations or TRANSLATIONS)
        }

        #######################################################################

        self.auth_table = auth_table
        self.site_name = site_name
        self.forms = forms
        self.read_only = read_only
        self.form_config_map = {form.slug: form for form in self.forms}

        with open(os.path.join(ASSET_PATH, "index.html")) as f:
            self.template = f.read()

        #######################################################################

        private_app = FastAPI(
            docs_url=None,
            redoc_url=None,
            debug=debug,
            exception_handlers={500: log_error},
        )
        private_app.mount("/docs/", swagger_ui(schema_url="../openapi.json"))

        for table_config in table_configs:
            table_class = table_config.table_class
            visible_column_names = table_config.get_visible_column_names()
            visible_filter_names = table_config.get_visible_filter_names()
            rich_text_columns_names = (
                table_config.get_rich_text_columns_names()
            )
            media_columns_names = table_config.get_media_columns_names()

            validators = table_config.validators
            if table_class in (auth_table, session_table):
                validators = validators or Validators()
                validators.every = [superuser_validators, *validators.every]

            FastAPIWrapper(
                root_url=f"/tables/{table_class._meta.tablename}/",
                fastapi_app=private_app,
                piccolo_crud=PiccoloCRUD(
                    table=table_class,
                    read_only=read_only,
                    page_size=page_size,
                    schema_extra={
                        "visible_column_names": visible_column_names,
                        "visible_filter_names": visible_filter_names,
                        "rich_text_columns": rich_text_columns_names,
                        "media_columns": media_columns_names,
                    },
                    validators=validators,
                    hooks=table_config.hooks,
                ),
                fastapi_kwargs=FastAPIKwargs(
                    all_routes={
                        "tags": [f"{table_class._meta.tablename.capitalize()}"]
                    },
                ),
            )

        private_app.add_api_route(
            path="/tables/",
            endpoint=self.get_table_list,  # type: ignore
            methods=["GET"],
            response_model=t.List[str],
            tags=["Tables"],
        )

        private_app.add_api_route(
            path="/forms/",
            endpoint=self.get_forms,  # type: ignore
            methods=["GET"],
            tags=["Forms"],
            response_model=t.List[FormConfigResponseModel],
        )

        private_app.add_api_route(
            path="/forms/{form_slug:str}/",
            endpoint=self.get_single_form,  # type: ignore
            methods=["GET"],
            tags=["Forms"],
        )

        private_app.add_api_route(
            path="/forms/{form_slug:str}/schema/",
            endpoint=self.get_single_form_schema,  # type: ignore
            methods=["GET"],
            tags=["Forms"],
        )

        private_app.add_api_route(
            path="/forms/{form_slug:str}/",
            endpoint=self.post_single_form,  # type: ignore
            methods=["POST"],
            tags=["Forms"],
        )

        private_app.add_api_route(
            path="/user/",
            endpoint=self.get_user,  # type: ignore
            methods=["GET"],
            tags=["User"],
            response_model=UserResponseModel,
        )

        private_app.add_route(
            path="/change-password/",
            route=change_password(
                login_url="./../../public/login/",
                session_table=session_table,
                read_only=read_only,
            ),
            methods=["POST"],
        )

        #######################################################################
        # Media

        private_app.add_api_route(
            path="/media/",
            endpoint=self.store_file,  # type: ignore
            methods=["POST"],
            tags=["Media"],
            response_model=StoreFileResponseModel,
        )

        private_app.add_api_route(
            path="/media/generate-file-url/",
            endpoint=self.generate_file_url,  # type: ignore
            methods=["POST"],
            tags=["Media"],
            response_model=GenerateFileURLResponseModel,
        )

        for table_config in self.table_configs:
            if table_config.media_columns:
                for (
                    column,
                    media_storage,
                ) in table_config.media_columns.items():
                    if isinstance(media_storage, LocalMediaStorage):
                        private_app.mount(
                            path=f"/media-files/{column._meta.table._meta.tablename}/{column._meta.name}/",  # noqa: E501
                            app=StaticFiles(
                                directory=media_storage.media_path
                            ),
                        )

        #######################################################################

        public_app = FastAPI(
            redoc_url=None,
            docs_url=None,
            debug=debug,
            exception_handlers={500: log_error},
        )
        public_app.mount("/docs/", swagger_ui(schema_url="../openapi.json"))

        if not rate_limit_provider:
            rate_limit_provider = InMemoryLimitProvider(
                limit=1000, timespan=300
            )

        public_app.mount(
            path="/login/",
            app=RateLimitingMiddleware(
                app=session_login(
                    auth_table=self.auth_table,
                    session_table=session_table,
                    session_expiry=session_expiry,
                    max_session_expiry=max_session_expiry,
                    redirect_to=None,
                    production=production,
                ),
                provider=rate_limit_provider,
            ),
        )

        public_app.add_route(
            path="/logout/",
            route=session_logout(session_table=session_table),
            methods=["POST"],
        )

        # We make the meta endpoint available without auth, because it contains
        # the site name.
        public_app.add_api_route(
            "/meta/", endpoint=self.get_meta, tags=["Meta"]  # type: ignore
        )

        # The translations are public, because we need them on the login page.
        public_app.add_api_route(
            "/translations/",
            endpoint=self.get_translation_list,  # type: ignore
            methods=["GET"],
            tags=["Translations"],
            response_model=TranslationListResponse,
        )

        public_app.add_api_route(
            "/translations/{language_code:str}/",
            endpoint=self.get_translation,  # type: ignore
            methods=["GET"],
            tags=["Translations"],
            response_model=Translation,
        )

        #######################################################################

        self.router.add_route(
            path="/", endpoint=self.get_root, methods=["GET"]
        )

        self.mount(
            path="/css",
            app=StaticFiles(directory=os.path.join(ASSET_PATH, "css")),
        )

        self.mount(
            path="/js",
            app=StaticFiles(directory=os.path.join(ASSET_PATH, "js")),
        )

        auth_middleware = partial(
            AuthenticationMiddleware,
            backend=SessionsAuthBackend(
                auth_table=auth_table,
                session_table=session_table,
                admin_only=True,
                increase_expiry=increase_expiry,
            ),
            on_error=handle_auth_exception,
        )

        self.mount(path="/api", app=auth_middleware(private_app))
        self.mount(path="/public", app=public_app)

    async def get_root(self, request: Request) -> HTMLResponse:
        return HTMLResponse(self.template)

    ###########################################################################

    def _get_media_storage(
        self, table_name: str, column_name: str
    ) -> MediaStorage:
        """
        Retrieve the ``MediaStorage`` for the given column.

        :raises HTTPException:
            If a matching ``MediaStorage`` can't be found.

        """
        table_config = self.table_config_map.get(table_name)
        if not table_config:
            raise HTTPException(status_code=404, detail="No such table found.")

        media_columns = table_config.media_columns

        if media_columns is None:
            raise HTTPException(
                status_code=422,
                detail="No media columns are configured for this table.",
            )

        try:
            column = table_config.table_class._meta.get_column_by_name(
                column_name
            )
        except ValueError:
            raise HTTPException(
                status_code=404,
                detail="No such column found.",
            )

        media_storage = media_columns.get(column)

        if not media_storage:
            raise HTTPException(
                status_code=422,
                detail="This column is not configured as a media_column.",
            )

        return media_storage

    async def store_file(
        self,
        request: Request,
        table_name: str = Form(None),
        column_name: str = Form(None),
        file: UploadFile = File(...),
    ) -> StoreFileResponseModel:
        """
        Stores in the file in the configured ``MediaStorage``, and returns a
        unique key for identifying that file.
        """
        if self.read_only:
            raise HTTPException(
                status_code=405, detail="Running in read-only mode."
            )

        media_storage = self._get_media_storage(
            table_name=table_name, column_name=column_name
        )

        try:
            file_key = await media_storage.store_file(
                file_name=file.filename,
                file=file.file,
                user=request.user.user,
            )
        except ValueError as exception:
            raise HTTPException(status_code=422, detail=str(exception))
        return StoreFileResponseModel(file_key=file_key)

    async def generate_file_url(
        self, request: Request, model: GenerateFileURLRequestModel
    ) -> GenerateFileURLResponseModel:
        """
        Returns a URL for accessing the given file.

        We don't use a GET for this endpoint, as using a GET param to pass the
        ``file_key`` is too restrictive on which characters can be used.
        """
        if self.read_only:
            raise HTTPException(
                status_code=405, detail="Running in read-only mode."
            )

        media_storage = self._get_media_storage(
            table_name=model.table_name, column_name=model.column_name
        )
        file_url = await media_storage.generate_file_url(
            file_key=model.file_key,
            root_url=(
                f"./api/media-files/{model.table_name}"
                f"/{model.column_name}/"
            ),
            user=request.user.user,
        )
        return GenerateFileURLResponseModel(file_url=file_url)

    ###########################################################################

    def get_user(self, request: Request) -> UserResponseModel:
        return UserResponseModel(
            username=request.user.display_name,
            user_id=str(request.user.user_id),
        )

    ###########################################################################
    # Custom forms

    def get_forms(self) -> t.List[FormConfigResponseModel]:
        """
        Returns a list of all forms registered with the admin.
        """
        return [
            FormConfigResponseModel(
                name=form.name, slug=form.slug, description=form.description
            )
            for form in self.forms
        ]

    def get_single_form(self, form_slug: str) -> FormConfigResponseModel:
        """
        Returns the FormConfig for the given form.
        """
        form = self.form_config_map.get(form_slug, None)
        if form is None:
            raise HTTPException(status_code=404, detail="No such form found")
        else:
            return FormConfigResponseModel(
                name=form.name,
                slug=form.slug,
                description=form.description,
            )

    def get_single_form_schema(self, form_slug: str) -> t.Dict[str, t.Any]:
        form_config = self.form_config_map.get(form_slug)

        if form_config is None:
            raise HTTPException(status_code=404, detail="No such form found")
        else:
            return form_config.pydantic_model.schema()

    async def post_single_form(
        self, request: Request, form_slug: str
    ) -> JSONResponse:
        """
        Handles posting of custom forms.
        """
        form_config = self.form_config_map.get(form_slug)
        data = await request.json()

        if form_config is None:
            raise HTTPException(status_code=404, detail="No such form found")

        try:
            model_instance = form_config.pydantic_model(**data)
        except ValidationError as exception:
            # We use 'detail' as it mirrors what FastAPI returns for Pydantic
            # errors, allowing us to use the same error display logic in the
            # front end.
            return JSONResponse(
                {"detail": json.loads(exception.json())}, status_code=422
            )

        try:
            endpoint = form_config.endpoint
            if inspect.iscoroutinefunction(endpoint):
                response = await endpoint(  # type: ignore
                    request, model_instance
                )
            else:
                response = endpoint(request, model_instance)
        except ValueError as exception:
            return JSONResponse(
                {"custom_form_error": str(exception)}, status_code=422
            )

        message = (
            response if isinstance(response, str) else "Successfully submitted"
        )
        return JSONResponse({"custom_form_success": message})

    ###########################################################################

    def get_meta(self) -> MetaResponseModel:
        return MetaResponseModel(
            piccolo_admin_version=PICCOLO_ADMIN_VERSION,
            site_name=self.site_name,
        )

    ###########################################################################

    def get_table_list(self) -> t.List[str]:
        """
        Returns a list of all tables registered with the admin.
        """
        return [i.table_class._meta.tablename for i in self.table_configs]

    ###########################################################################

    def get_translation_list(self) -> TranslationListResponse:
        """
        Return a list of language codes and names for each available
        translation.
        """
        return TranslationListResponse(
            translations=[
                TranslationListItem(
                    language_code=translation.language_code,
                    language_name=translation.language_name,
                )
                for translation in self.translations_map.values()
            ],
            default_language_code=self.default_language_code,
        )

    def get_translation(self, language_code: str = "en") -> Translation:
        """
        Return a single language. The ``language_code`` is an IETF language
        code, for example 'en' for English.
        """
        translation = self.translations_map.get(language_code.lower())
        if translation is None:
            raise HTTPException(
                status_code=404, detail="Translation not found"
            )
        return translation


def get_all_tables(
    tables: t.Sequence[t.Type[Table]],
) -> t.Sequence[t.Type[Table]]:
    """
    Fetch any related tables, and include them.
    """
    output: t.List[t.Type[Table]] = []

    def get_references(table: t.Type[Table]):
        references: t.List[t.Union[t.Type[Table], t.Any]] = [
            i._foreign_key_meta.references
            for i in table._meta.foreign_key_columns
        ]
        for reference in references:
            table = (
                reference.resolve()
                if isinstance(reference, LazyTableReference)
                else reference
            )

            if table not in output:
                output.append(table)
                get_references(table)

    for table in tables:
        if table not in output:
            output.append(table)
        get_references(table)

    return output


def create_admin(
    tables: t.Sequence[t.Union[t.Type[Table], TableConfig]],
    forms: t.List[FormConfig] = [],
    auth_table: t.Optional[t.Type[BaseUser]] = None,
    session_table: t.Optional[t.Type[SessionsBase]] = None,
    session_expiry: timedelta = timedelta(hours=1),
    max_session_expiry: timedelta = timedelta(days=7),
    increase_expiry: t.Optional[timedelta] = timedelta(minutes=20),
    page_size: int = 15,
    read_only: bool = False,
    rate_limit_provider: t.Optional[RateLimitProvider] = None,
    production: bool = False,
    site_name: str = "Piccolo Admin",
    default_language_code: str = "auto",
    translations: t.List[Translation] = None,
    auto_include_related: bool = True,
    allowed_hosts: t.Sequence[str] = [],
    debug: bool = False,
):
    """
    :param tables:
        Each of the tables will be added to the admin.
    :param forms:
        For each :class:`FormConfig <piccolo_admin.endpoints.FormConfig>`
        specified, a form will automatically be rendered in the user interface,
        accessible via the sidebar.
    :param auth_table:
        Either a :class:`BaseUser <piccolo.apps.user.tables.BaseUser>`, or
        ``BaseUser`` subclass table, which is used for fetching users.
        Defaults to ``BaseUser`` if none if specified.
    :param session_table:
        Either a :class:`SessionsBase <piccolo_api.session_auth.tables.SessionsBase>`,
        or ``SessionsBase`` subclass table, which is used for storing and
        querying session tokens. Defaults to ``SessionsBase`` if none if
        specified.
    :param session_expiry:
        How long a session is valid for.
    :param max_session_expiry:
        The maximum time a session is valid for, taking into account any
        refreshes using ``increase_expiry``.
    :param increase_expiry:
        If set, the ``session_expiry`` will be increased by this amount if it's
        close to expiry.
    :param page_size:
        The admin API paginates content - this sets the default number of
        results on each page.
    :param read_only:
        If ``True``, all non auth endpoints only respond to GET requests - the
        admin can still be viewed, and the data can be filtered. Useful for
        creating online demos.
    :param rate_limit_provider:
        Rate limiting middleware is used to protect the login endpoint
        against brute force attack. If not set, an
        :class:`InMemoryLimitProvider <piccolo_api.rate_limiting.middleware.InMemoryLimitProvider>`
        will be configured with reasonable defaults.
    :param production:
        If ``True``, the admin will enforce stronger security - for example,
        the cookies used will be secure, meaning they are only sent over
        HTTPS.
    :param site_name:
        Specify a different site name in the admin UI (default
        ``'Piccolo Admin'``).
    :param default_language_code:
        Specify the default language used in the admin UI. The value should be
        an `IETF language tag <https://en.wikipedia.org/wiki/IETF_language_tag>`_,
        for example ``'en'`` for English. To see available values see
        ``piccolo_admin/translations/data.py``. The UI will be automatically
        translated into this language. If a value of ``'auto'`` is specified,
        then we check the user's browser for the language they prefer, using
        the ``navigator.language`` JavaScript API.
    :param translations:
        Specify which translations are available. By default, we use every
        translation in ``piccolo_admin/translations/data.py``.

        Here's an example - if we know our users only speak English or
        Croatian, we can specify that only those translations are visible
        in the language selector in the UI::

            from piccolo.translations.data import ENGLISH, CROATIAN

            create_admin(
                tables=[TableA, TableB],
                default_language_code='hr',
                translations=[ENGLISH, CROATIAN]
            )

        You can also use this to provide your own translations, if there's a
        language we don't currently support (though please open a PR to add
        it!)::

            from piccolo.translations.models import Translation
            from piccolo.translations.data import ENGLISH

            MY_LANGUAGE = Translation(
                language_code='xx',
                language_name='My Language',
                translations={
                    'Welcome': 'XXXXX',
                    ...
                }
            )

            create_admin(
                tables=[TableA, TableB],
                default_language_code='xx',
                translations=[ENGLISH, MY_LANGUAGE]
            )

    :param auto_include_related:
        If a table has foreign keys to other tables, those tables will also be
        included in the admin by default, if not already specified. Otherwise
        the admin won't work as expected.
    :param allowed_hosts:
        This is used by the :class:`CSRFMiddleware <piccolo_api.csrf.middleware.CSRFMiddleware>`
        as an additional layer of protection when the admin is run under HTTPS.
        It must be a sequence of strings, such as ``['my_site.com']``.
    :param debug:
        If ``True``, debug mode is enabled. Any unhandled exceptions will
        return a stack trace, rather than a generic 500 error. Don't use this
        in production!

    """  # noqa: E501
    auth_table = auth_table or BaseUser
    session_table = session_table or SessionsBase

    if auto_include_related:
        table_config_map: t.Dict[t.Type[Table], t.Optional[TableConfig]] = {}

        for i in tables:
            if isinstance(i, TableConfig):
                table_config_map[i.table_class] = i
            else:
                table_config_map[i] = None

        all_table_classes = get_all_tables(tuple(table_config_map.keys()))

        all_table_classes_with_configs: t.List[
            t.Union[t.Type[Table], TableConfig]
        ] = []
        for i in all_table_classes:
            table_config = table_config_map.get(i)
            if table_config:
                all_table_classes_with_configs.append(table_config)
            else:
                all_table_classes_with_configs.append(i)

        tables = all_table_classes_with_configs

    return AdminRouter(
        *tables,
        forms=forms,
        auth_table=auth_table,
        session_table=session_table,
        session_expiry=session_expiry,
        max_session_expiry=max_session_expiry,
        increase_expiry=increase_expiry,
        page_size=page_size,
        read_only=read_only,
        rate_limit_provider=rate_limit_provider,
        production=production,
        site_name=site_name,
        default_language_code=default_language_code,
        translations=translations,
        allowed_hosts=allowed_hosts,
        debug=debug,
    )



"""
Runs the admin in read only mode - useful for letting people evaluate the admin
online without risk of abuse.
"""
import targ
import uvicorn

from piccolo_admin.example import (
    Director,
    Movie,
    Sessions,
    User,
    create_admin,
    create_schema,
    populate_data,
    set_engine,
)

APP = create_admin(
    [Director, Movie],
    auth_table=User,
    session_table=Sessions,
    read_only=True,
)


def sandbox(host: str = "localhost", port: int = 8080):
    """
    Run the Piccolo Admin in read only mode with a SQLite database.

    :param host:
        Which host to serve the app on.
    :param host:
        Which port to serve the app on.

    """
    set_engine()
    create_schema(persist=False)
    populate_data()

    uvicorn.run(APP, host=host, port=port)


if __name__ == "__main__":
    cli = targ.CLI()
    cli.register(sandbox)
    cli.run(solo=True)



import datetime
import typing as t
from decimal import Decimal

DIRECTORS: t.List[t.Dict[str, t.Any]] = [
    {
        "id": 1,
        "name": "Peter Jackson",
        "years_nominated": [2002, 2004],
        "gender": "m",
    },
    {
        "id": 2,
        "name": "George Lucas",
        "years_nominated": [1974, 1978],
        "gender": "m",
    },
    {"id": 3, "name": "J.J. Abrams", "years_nominated": [], "gender": "m"},
    {"id": 4, "name": "Gareth Edwards", "years_nominated": [], "gender": "m"},
    {"id": 5, "name": "Irvin Kershner", "years_nominated": [], "gender": "m"},
    {
        "id": 6,
        "name": "Richard Marquand",
        "years_nominated": [],
        "gender": "m",
    },
    {"id": 7, "name": "Rian Johnson", "years_nominated": [], "gender": "m"},
    {
        "id": 8,
        "name": "Ron Howard",
        "years_nominated": [2002, 2009],
        "gender": "m",
    },
]


MOVIES: t.List[t.Dict[str, t.Any]] = [
    {
        "name": "Star Wars: Episode IV - A New Hope",
        "rating": 93.3,
        "duration": datetime.timedelta(minutes=125),
        "director": 2,
        "description": (
            "After Princess Leia, the leader of the Rebel Alliance, is held "
            "hostage by Darth Vader, Luke and Han Solo must free her and "
            "destroy the powerful weapon created by the Galactic Empire."
        ),
        "release_date": datetime.date(year=1977, month=12, day=27),
        "oscar_nominations": 11,
        "won_oscar": True,
        "box_office": Decimal("775.5"),
        "barcode": 3999361649,
        "genre": 2,
    },
    {
        "name": "The Empire Strikes Back",
        "rating": 94.0,
        "duration": datetime.timedelta(minutes=127),
        "director": 5,
        "description": (
            "Darth Vader is adamant about turning Luke Skywalker to the dark "
            "side. Master Yoda trains Luke to become a Jedi Knight while his "
            "friends try to fend off the Imperial fleet."
        ),
        "release_date": datetime.date(year=1980, month=5, day=20),
        "oscar_nominations": 4,
        "won_oscar": True,
        "box_office": Decimal("547.9"),
        "barcode": 2384023507,
        "genre": 2,
    },
    {
        "name": "Return of the Jedi",
        "rating": 82.0,
        "duration": datetime.timedelta(minutes=136),
        "director": 6,
        "description": (
            "Luke Skywalker attempts to bring his father back to the light "
            "side of the 'Force'. At the same time, the rebels hatch a plan "
            "to destroy the second Death Star."
        ),
        "release_date": datetime.date(year=1983, month=6, day=2),
        "oscar_nominations": 5,
        "won_oscar": True,
        "box_office": Decimal("475.3"),
        "barcode": 9874111507,
        "genre": 2,
    },
    {
        "name": "Star Wars: Episode I – The Phantom Menace",
        "rating": 53.0,
        "duration": datetime.timedelta(minutes=136),
        "director": 2,
        "description": (
            "Two Jedi Knights set out to search for someone who can "
            "bring peace to the Force. Their search ends when they come "
            "across a young, gifted boy. But the Sith returns to stake claim "
            "to the Force."
        ),
        "release_date": datetime.date(year=1999, month=7, day=16),
        "oscar_nominations": 3,
        "won_oscar": False,
        "box_office": Decimal("1027.0"),
        "barcode": 4637581510,
        "genre": 2,
    },
    {
        "name": "Star Wars: Episode II – Attack of the Clones",
        "rating": 65,
        "duration": datetime.timedelta(minutes=144),
        "director": 2,
        "description": (
            "While pursuing an assassin, Obi Wan uncovers a sinister plot to "
            "destroy the Republic. With the fate of the galaxy hanging in the "
            "balance, the Jedi must defend the galaxy against the evil Sith."
        ),
        "release_date": datetime.date(year=2002, month=5, day=16),
        "oscar_nominations": 1,
        "won_oscar": False,
        "box_office": Decimal("653.0"),
        "barcode": 8392825607,
        "genre": 2,
    },
    {
        "name": "Star Wars: Episode III – Revenge of the Sith",
        "rating": 80.0,
        "duration": datetime.timedelta(minutes=140),
        "director": 2,
        "description": (
            "Anakin joins forces with Obi-Wan and sets Palpatine free from "
            "the evil clutches of Count Doku. But he falls prey to mind games "
            "that are played by Palpatine and the Jedis and gives into "
            "temptation."
        ),
        "release_date": datetime.date(year=2005, month=5, day=19),
        "oscar_nominations": 1,
        "won_oscar": False,
        "box_office": Decimal("868.4"),
        "barcode": 5174272602,
        "genre": 2,
    },
    {
        "name": "Star Wars: The Force Awakens",
        "rating": 93.0,
        "duration": datetime.timedelta(minutes=135),
        "director": 3,
        "description": (
            "A new order threatens to destroy the New Republic. Finn, Rey and "
            "Poe, backed by the Resistance and the Republic, must find a way "
            "to stop them and find Luke, the last surviving Jedi."
        ),
        "release_date": datetime.date(year=2015, month=12, day=17),
        "oscar_nominations": 5,
        "won_oscar": False,
        "box_office": Decimal("2068.0"),
        "barcode": 9476361634,
        "genre": 2,
    },
    {
        "name": "Rogue One: A Star Wars Story",
        "rating": 83.1,
        "duration": datetime.timedelta(minutes=133),
        "director": 4,
        "description": (
            "When she's a child, Jyn's father is forcibly taken by the Empire "
            "to help them complete the Death Star. As an adult, she joins a "
            "group of resistance fighters, who aim to steal its blueprints."
        ),
        "release_date": datetime.date(year=2016, month=12, day=13),
        "oscar_nominations": 2,
        "won_oscar": False,
        "box_office": Decimal("1056.0"),
        "barcode": 6396501778,
        "genre": 2,
    },
    {
        "name": "Star Wars: The Last Jedi",
        "rating": 91.2,
        "duration": datetime.timedelta(minutes=152),
        "director": 7,
        "description": (
            "Rey seeks to learn the ways of the Jedi under Luke Skywalker, "
            "its remaining member, to reinvigorate the Resistance's war "
            "against the First Order."
        ),
        "release_date": datetime.date(year=2017, month=12, day=14),
        "oscar_nominations": 4,
        "won_oscar": False,
        "box_office": Decimal("1333.0"),
        "barcode": 1020122875,
        "genre": 2,
    },
    {
        "name": "Solo: A Star Wars Story",
        "rating": 70.5,
        "duration": datetime.timedelta(minutes=135),
        "director": 8,
        "description": (
            "In a galaxy where hyperfuel is in high demand, Han Solo gets "
            "involved in a large-scale heist within the criminal underworld "
            "and meets individuals who change his life."
        ),
        "release_date": datetime.date(year=2018, month=5, day=25),
        "oscar_nominations": 0,
        "won_oscar": False,
        "box_office": Decimal("393.2"),
        "barcode": 4949091770,
        "genre": 2,
    },
    {
        "name": "Star Wars: The Rise of Skywalker",
        "rating": 54.7,
        "duration": datetime.timedelta(minutes=142),
        "director": 3,
        "description": (
            "The surviving Resistance faces the First Order once more as Rey, "
            "Finn and Poe Dameron's journey continues. With the power and "
            "knowledge of generations behind them, the final battle commences."
        ),
        "release_date": datetime.date(year=2019, month=12, day=19),
        "oscar_nominations": 3,
        "won_oscar": False,
        "box_office": Decimal("1074.0"),
        "barcode": 3592641903,
        "genre": 2,
    },
    {
        "name": "The Lord of the Rings: The Fellowship of the Ring",
        "rating": 91.1,
        "duration": datetime.timedelta(minutes=208),
        "director": 1,
        "description": (
            "A young hobbit, Frodo, who has found the One Ring that belongs "
            "to the Dark Lord Sauron, begins his journey with eight "
            "companions to Mount Doom, the only place where it can be "
            "destroyed."
        ),
        "release_date": datetime.date(year=2001, month=12, day=10),
        "oscar_nominations": 12,
        "won_oscar": True,
        "box_office": Decimal("887.8"),
        "barcode": 9520764061,
        "genre": 1,
    },
    {
        "name": "The Lord of the Rings: The Two Towers",
        "rating": 95.2,
        "duration": datetime.timedelta(minutes=223),
        "director": 1,
        "description": (
            "Frodo and Sam arrive in Mordor with the help of Gollum. A number "
            "of new allies join their former companions to defend Isengard as "
            "Saruman launches an assault on it."
        ),
        "release_date": datetime.date(year=2002, month=12, day=11),
        "oscar_nominations": 6,
        "won_oscar": True,
        "box_office": Decimal("951.2"),
        "barcode": 2545685641,
        "genre": 1,
    },
    {
        "name": "The Lord of the Rings: The Return of the King",
        "rating": 93.0,
        "duration": datetime.timedelta(minutes=201),
        "director": 1,
        "description": (
            "The former Fellowship members prepare for the final battle. "
            "While Frodo and Sam approach Mount Doom to destroy the One Ring, "
            "they follow Gollum unaware of the path he is leading them to."
        ),
        "release_date": datetime.date(year=2003, month=12, day=17),
        "oscar_nominations": 12,
        "won_oscar": True,
        "box_office": Decimal("1142.0"),
        "barcode": 6118332538,
        "genre": 1,
    },
    {
        "name": "The Hobbit: An Unexpected Journey",
        "rating": 64.2,
        "duration": datetime.timedelta(minutes=182),
        "director": 1,
        "description": (
            "Bilbo Baggins, a hobbit, is persuaded into accompanying a wizard "
            "and a group of dwarves on a journey to reclaim the city of "
            "Erebor and all its riches from the dragon Smaug."
        ),
        "release_date": datetime.date(year=2012, month=12, day=12),
        "oscar_nominations": 3,
        "won_oscar": False,
        "box_office": Decimal("1021.0"),
        "barcode": 3604722656,
        "genre": 1,
    },
    {
        "name": "The Hobbit: The Desolation of Smaug",
        "rating": 74.9,
        "duration": datetime.timedelta(minutes=187),
        "director": 1,
        "description": (
            "Bilbo Baggins, a hobbit, and his companions face great dangers "
            "on their journey to Laketown. Soon, they reach the Lonely "
            "Mountain, where Bilbo comes face-to-face with the fearsome "
            "dragon Smaug."
        ),
        "release_date": datetime.date(year=2013, month=12, day=13),
        "oscar_nominations": 3,
        "won_oscar": False,
        "box_office": Decimal("958.0"),
        "barcode": 2522965834,
        "genre": 1,
    },
    {
        "name": "The Hobbit: The Battle of the Five Armies",
        "rating": 59.0,
        "duration": datetime.timedelta(minutes=164),
        "director": 1,
        "description": (
            "Bilbo fights against a number of enemies to save the life of his "
            "Dwarf friends and protects the Lonely Mountain after a conflict "
            "arises."
        ),
        "release_date": datetime.date(year=2014, month=12, day=1),
        "oscar_nominations": 1,
        "won_oscar": False,
        "box_office": Decimal("956.0"),
        "barcode": 3939605744,
        "genre": 2,
    },
]


STUDIOS: t.List[t.Dict[str, t.Any]] = [
    {
        "name": "Shepperton Studios",
        "facilities": {
            "sound_stage": True,
            "catering": True,
            "water_tank": True,
        },
    }
]


TICKETS: t.List[t.Dict[str, t.Any]] = [
    {
        "booked_by": "star_wars_super_fan_1",
        "movie": 1,
        "start_date": datetime.date(year=2022, month=1, day=20),
        "start_time": datetime.time(hour=16, minute=15),
        "booked_on": datetime.datetime(
            year=2022, month=1, day=12, hour=9, minute=30
        ),
    },
    {
        "booked_by": "bilbo_63",
        "movie": 15,
        "start_date": datetime.date(year=2022, month=2, day=22),
        "start_time": datetime.time(hour=19, minute=30),
        "booked_on": datetime.datetime(
            year=2022, month=2, day=11, hour=16, minute=30
        ),
    },
]


# Some random words used to generate fake movie names.
MOVIE_WORDS: t.List[str] = [
    "Adventure",
    "Alien",
    "Armageddon",
    "Android",
    "Bridge",
    "Captain",
    "Destruction",
    "Evolution",
    "Extreme",
    "Fantasy",
    "Friend",
    "Galaxy",
    "King",
    "Lord",
    "Master",
    "Max",
    "Mutant",
    "Outpost",
    "Partner",
    "Phantom",
    "Power",
    "Returns",
    "Revenge",
    "Robot",
    "Star",
    "Throne",
    "Thunder",
    "Twin",
]



import typing as t

from pydantic import BaseModel, Field


class TranslationListItem(BaseModel):
    language_name: str = Field(description="e.g. 'English'")
    language_code: str = Field(description="e.g. 'en'")


class TranslationListResponse(BaseModel):
    translations: t.List[TranslationListItem]
    default_language_code: str = Field(description="e.g. 'en'")


class Translation(BaseModel):
    """
    Used to provide translations in the UI.

    :param language_name:
        A human readable representation of the language. For example
        ``'English'``.
    :param language_code:
        The `IETF language tag <https://en.wikipedia.org/wiki/IETF_language_tag>`_.
        For English it is ``'en'``. However, it also allows us to specify
        dialects like ``'en-US'`` for American English, or ``'en-GB'`` for
        British English.
    :param translations:
        A mapping of English words / phrases to their translated form. For
        example::

            Translation(
                language_name='Portuguese',
                language_code='pt',
                translations={
                    'Welcome to': 'Bem-vindo ao',
                    ...
                }
            )

    """  # noqa: E501

    language_name: str
    language_code: str
    translations: t.Dict[str, str]



"""
See https://github.com/piccolo-orm/piccolo_admin/issues/173 for more info.

NOTE: Flake8's line length warning has been disabled for this file - see
.flake8, so we don't have to worry about wrapping long translations.

NOTE: To validate the translations (to see if any are missing)::

    # At the root of the project
    python scripts/get_translations.py validate

NOTE: One of the fastest ways to get the translations in bulk is using Google
Spreadsheets. For example, this function translates a value from English to
Portuguese `=GOOGLETRANSLATE(A1,"en","pt")`. A CSV file can be downloaded, and
then converted to JSON using a tool like this
https://www.convertcsv.com/csv-to-json.htm.

"""

import typing as t

from piccolo_admin.translations.models import Translation

# For now there aren't any words which are different between dialects, so we
# only need one form of English.
ENGLISH = Translation(
    language_name="English",
    language_code="en",
    translations={
        "About": "About",
        "Add Row": "Add Row",
        "Add": "Add",
        "Apply": "Apply",
        "Ascending": "Ascending",
        "Back to home page": "Back to home page",
        "Back": "Back",
        "Change password": "Change password",
        "Clear filters": "Clear filters",
        "Close": "Close",
        "Create": "Create",
        "Current password": "Current password",
        "Dark Mode": "Dark Mode",
        "Days": "Days",
        "Delete": "Delete",
        "Descending": "Descending",
        "Edit": "Edit",
        "Export CSV": "Export CSV",
        "Filter": "Filter",
        "Form submitted": "Form submitted",
        "Forms": "Forms",
        "Go to page": "Go to page",
        "Hide Filters": "Hide Filters",
        "Hide referencing tables": "Hide referencing tables",
        "Home": "Home",
        "Hours": "Hours",
        "Light Mode": "Light Mode",
        "Loading": "Loading",
        "Log out": "Log out",
        "Login": "Login",
        "Minutes": "Minutes",
        "New password confirmation": "New password confirmation",
        "New password": "New password",
        "New value": "New value",
        "No results found": "No results found",
        "of": "of",
        "page": "page",
        "Password": "Password",
        "result(s)": "result(s)",
        "rows": "rows",
        "Save": "Save",
        "Seconds": "Seconds",
        "Select a column to update": "Select a column to update",
        "Select a Column": "Select a Column",
        "Select a table in the sidebar to get started.": "Select a table in the sidebar to get started.",
        "selected result(s) on": "selected result(s) on",
        "Show Filters": "Show filters",
        "Show referencing tables": "Show referencing tables",
        "Showing": "Showing",
        "Sort by": "Sort by",
        "Sort": "Sort",
        "Submit": "Submit",
        "Tables": "Tables",
        "Thanks for using Piccolo Admin.": "Thanks for using Piccolo Admin.",
        "Update": "Update",
        "Use again": "Use again",
        "Username": "Username",
        "Version": "Version",
        "Weeks": "Weeks",
        "Welcome to": "Welcome to",
        "with a matching": "with a matching",
    },
)


WELSH = Translation(
    language_name="Welsh",
    language_code="cy",
    translations={
        "About": "Am",
        "Add Row": "Ychwanegu Rhes",
        "Add": "Ychwanegu",
        "Apply": "Ymgeisiwch",
        "Ascending": "Esgynnol",
        "Back to home page": "Yn ôl i'r dudalen gartref",
        "Back": "Ol",
        "Change password": "Newid cyfrinair",
        "Clear filters": "Clirio hidlwyr",
        "Close": "Cau",
        "Create": "Creu",
        "Current password": "Cyfrinair cyfredol",
        "Dark Mode": "Modd Yywyll",
        "Days": "Dyddiau",
        "Delete": "Dileu",
        "Descending": "Disgyn",
        "Edit": "Golygu",
        "Export CSV": "Allforio CSV",
        "Filter": "Ffilter",
        "Form submitted": "Ffurflen wedi'i chyflwyno",
        "Forms": "Ffurflenni",
        "Go to page": "Ewch i'r dudalen",
        "Hide Filters": "Cuddio hidlwyr",
        "Hide referencing tables": "Cuddio tablau cyfeirio",
        "Home": "Cartref",
        "Hours": "Oriau",
        "Light Mode": "Modd Golau",
        "Loading": "Llwytho",
        "Log out": "Allgofnodi",
        "Login": "Mewngofnodi",
        "Minutes": "Munudau",
        "New password confirmation": "Cadarnhad cyfrinair newydd",
        "New password": "Cyfrinair newydd",
        "New value": "Gwerth newydd",
        "No results found": "Heb ganfod canlyniadau",
        "of": "o",
        "page": "tudalen",
        "Password": "Cyfrinair",
        "result(s)": "canlyniad(au)",
        "rows": "rhesi",
        "Save": "Arbed",
        "Seconds": "Eiliadau",
        "Select a column to update": "Dewiswch golofn i'w diweddaru",
        "Select a Column": "Dewiswch Golofn",
        "Select a table in the sidebar to get started.": "Dewiswch un o'r tablau yn y bar ochr i ddechrau.",
        "selected result(s) on": "canlyniad(au) dethol ymlaen",
        "Show Filters": "Dangos hidlwyr",
        "Show referencing tables": "Dangos tablau cyfeirio",
        "Showing": "Yn dangos",
        "Sort by": "Trefnu yn ôl",
        "Sort": "Didoli",
        "Submit": "Cyflwyno",
        "Tables": "Tablau",
        "Thanks for using Piccolo Admin.": "Diolch am ddefnyddio Piccolo Admin.",
        "Update": "Diweddariad",
        "Use again": "Defnyddiwch eto",
        "Username": "Enw defnyddiwr",
        "Version": "Fersiwn",
        "Weeks": "Wythnosau",
        "Welcome to": "Croeso i",
        "with a matching": "gyda chyfateb",
    },
)

CROATIAN = Translation(
    language_name="Croatian",
    language_code="hr",
    translations={
        "About": "O",
        "Add Row": "Dodaj redak",
        "Add": "Dodaj",
        "Apply": "Primijeni",
        "Ascending": "Uzlazno",
        "Back to home page": "Vrati se na početnu stranicu",
        "Back": "Natrag",
        "Change password": "Promijeni lozinku",
        "Clear filters": "Obriši filtere",
        "Close": "Zatvori",
        "Create": "Kreiraj",
        "Current password": "Trenutna lozinka",
        "Dark Mode": "Tamni način rada",
        "Days": "Dani",
        "Delete": "Izbriši",
        "Descending": "Silazno",
        "Edit": "Uredi",
        "Export CSV": "Izvezi CSV",
        "Filter": "Filtar",
        "Form submitted": "Obrazac poslan",
        "Forms": "Forme",
        "Go to page": "Idi na stranicu",
        "Hide Filters": "Sakrij filtre",
        "Hide referencing tables": "Sakrij referentne tablice",
        "Home": "Početna",
        "Hours": "Sati",
        "Light Mode": "Svijetli način rada",
        "Loading": "Učitavanje",
        "Log out": "Odjava",
        "Login": "Prijaviti se",
        "Minutes": "Minute",
        "New password confirmation": "Potvrda nove lozinke",
        "New password": "Nova lozinka",
        "New value": "Nova vrijednost:",
        "No results found": "Nema rezultata",
        "of": "od",
        "page": "stranici",
        "Password": "Zaporka",
        "result(s)": "rezultat(a)",
        "rows": "redaka",
        "Save": "Spremi",
        "Seconds": "Sekunde",
        "Select a column to update": "Odaberite stupac za ažuriranje",
        "Select a Column": "Odaberite stupac",
        "Select a table in the sidebar to get started.": "Za početak odaberite jednu od tablica na bočnoj traci.",
        "selected result(s) on": "odabranih rezultat(a) na",
        "Show Filters": "Prikaži filtere",
        "Show referencing tables": "Prikaži referentne tablice",
        "Showing": "Pokazuje",
        "Sort by": "Sortiraj po",
        "Sort": "Sortiraj",
        "Submit": "Podnijeti",
        "Tables": "Tablice",
        "Thanks for using Piccolo Admin.": "Hvala što koristite Piccolo Admin.",
        "Update": "Ažuriraj",
        "Use again": "Koristi ponovno",
        "Username": "Korisničko ime",
        "Version": "Verzija",
        "Weeks": "Tjedni",
        "Welcome to": "Dobrodošli u",
        "with a matching": "s odgovarajućom kolumnom",
    },
)


PORTUGUESE = Translation(
    language_name="Portuguese",
    language_code="pt",
    translations={
        "About": "Sobre",
        "Add Row": "Adicionar linha",
        "Add": "Adicionar",
        "Apply": "Aplicar",
        "Ascending": "Ascendente",
        "Back to home page": "Voltar à página inicial",
        "Back": "Voltar atrás",
        "Change password": "Mudar senha",
        "Clear filters": "Limpar Filtros",
        "Close": "Fechar",
        "Create": "Criar",
        "Current password": "Senha atual",
        "Dark Mode": "Modo escuro",
        "Days": "Dias",
        "Delete": "Eliminar",
        "Descending": "descendente",
        "Edit": "Editar",
        "Export CSV": "Exportar CSV",
        "Filter": "Filtro",
        "Form submitted": "Formulário enviado",
        "Forms": "Formulários",
        "Go to page": "Ir para página",
        "Hide Filters": "Ocultar filtros",
        "Hide referencing tables": "Ocultar tabelas de referência",
        "Home": "Página inicial",
        "Hours": "Horas",
        "Light Mode": "Modo claro",
        "Loading": "Carregando",
        "Log out": "Sair",
        "Login": "Conecte-se",
        "Minutes": "Minutos",
        "New password confirmation": "Nova confirmação de senha",
        "New password": "Nova Senha",
        "New value": "Novo valor",
        "No results found": "Nenhum resultado encontrado",
        "of": "do",
        "page": "página",
        "Password": "Senha",
        "result(s)": "resultado(s)",
        "rows": "linhas",
        "Save": "Guardar",
        "Seconds": "Segundos",
        "Select a column to update": "Selecione uma coluna para atualizar",
        "Select a Column": "Selecione uma coluna",
        "Select a table in the sidebar to get started.": "Selecione uma tabela na barra lateral para começar.",
        "selected result(s) on": "Resultados selecionados (s) em",
        "Show Filters": "Mostrar filtros",
        "Show referencing tables": "Mostrar tabelas de referência",
        "Showing": "Mostrando",
        "Sort by": "Ordenar por",
        "Sort": "Ordenar",
        "Submit": "Enviar",
        "Tables": "Tabelas",
        "Thanks for using Piccolo Admin.": "Obrigado por usar Piccolo Admin.",
        "Update": "Atualizar",
        "Use again": "Use novamente",
        "Username": "Nome de usuário",
        "Version": "Versão",
        "Weeks": "Semanas",
        "Welcome to": "Bem-vindo ao",
        "with a matching": "com uma correspondência",
    },
)


GERMAN = Translation(
    language_name="German",
    language_code="de",
    translations={
        "About": "Über",
        "Add Row": "Zeile hinzufügen",
        "Add": "Hinzufügen",
        "Apply": "Anwenden",
        "Ascending": "Aufsteigend",
        "Back to home page": "Zurück zur Startseite",
        "Back": "Zurück",
        "Change password": "Passwort ändern",
        "Clear filters": "Filter löschen",
        "Close": "Schließen",
        "Create": "Anlegen",
        "Current password": "Jetziges Passwort",
        "Dark Mode": "Dunkler Modus",
        "Days": "Tage",
        "Delete": "Löschen",
        "Descending": "Absteigend",
        "Edit": "Bearbeiten",
        "Export CSV": "Exportieren nach CSV",
        "Filter": "Filter",
        "Form submitted": "Formular eingereicht",
        "Forms": "Formulare",
        "Go to page": "Gehen Sie zur Seite",
        "Hide Filters": "Filter ausblenden",
        "Hide referencing tables": "Referenzierungstabellen ausblenden",
        "Home": "Startseite",
        "Hours": "Std",
        "Light Mode": "Heller Modus",
        "Loading": "Wird geladen",
        "Log out": "Ausloggen",
        "Login": "Anmeldung",
        "Minutes": "Minuten",
        "New password confirmation": "Neues Passwort bestätigen",
        "New password": "Neues Passwort",
        "New value": "Neuer Wert",
        "No results found": "keine Ergebnisse gefunden",
        "of": "von",
        "page": "Seite",
        "Password": "Passwort",
        "result(s)": "Ergebnis(se)",
        "rows": "Zeilen",
        "Save": "Speichern",
        "Seconds": "Sekunden",
        "Select a column to update": "Wählen Sie eine Spalte aus, um zu aktualisieren",
        "Select a Column": "Wählen Sie eine Spalte aus",
        "Select a table in the sidebar to get started.": "Wählen Sie eine Tabelle in der Seitenleiste aus, um loszulegen.",
        "selected result(s) on": "ausgewählte(s) Ergebnis(se) auf",
        "Show Filters": "Filter anzeigen",
        "Show referencing tables": "Referenzierungstabellen anzeigen",
        "Showing": "Zeigen",
        "Sort by": "Sortieren nach",
        "Sort": "Sortieren",
        "Submit": "Abschicken",
        "Tables": "Tabellen",
        "Thanks for using Piccolo Admin.": "Vielen Dank für die Verwendung von Piccolo Admin.",
        "Update": "Aktualisieren",
        "Use again": "Wiederbenutzen",
        "Username": "Nutzername",
        "Version": "Version",
        "Weeks": "Wochen",
        "Welcome to": "Willkommen zu",
        "with a matching": "mit einem Matching",
    },
)


FRENCH = Translation(
    language_name="French",
    language_code="fr",
    translations={
        "About": "À propos de",
        "Add Row": "Ajouter une ligne",
        "Add": "Ajouter",
        "Apply": "Appliquer",
        "Ascending": "Croissant",
        "Back to home page": "Retour à la page d'accueil",
        "Back": "Retour",
        "Change password": "Changer le mot de passe",
        "Clear filters": "Supprimer les filtres",
        "Close": "Fermer",
        "Create": "Créer",
        "Current password": "Mot de passe actuel",
        "Dark Mode": "Mode sombre",
        "Days": "Journées",
        "Delete": "Effacer",
        "Descending": "Décroissant",
        "Edit": "Éditer",
        "Export CSV": "Exporter CSV",
        "Filter": "Filtre",
        "Form submitted": "Formulaire soumis",
        "Forms": "Formulaires",
        "Go to page": "Aller à la page",
        "Hide Filters": "Masquer les filtres",
        "Hide referencing tables": "Masquer les tables de référence",
        "Home": "Accueil",
        "Hours": "Heures",
        "Light Mode": "Mode léger",
        "Loading": "Chargement",
        "Log out": "Se déconnecter",
        "Login": "Connexion",
        "Minutes": "Minutes",
        "New password confirmation": "Confirmation du nouveau mot de passe",
        "New password": "Nouveau mot de passe",
        "New value": "Nouvelle valeur",
        "No results found": "Aucun résultat trouvé",
        "of": "de",
        "page": "page",
        "Password": "Mot de passe",
        "result(s)": "résultat(s)",
        "rows": "lignes",
        "Save": "Sauvegarder",
        "Seconds": "Secondes",
        "Select a column to update": "Sélectionnez une colonne à mettre à jour",
        "Select a Column": "Sélectionnez une colonne",
        "Select a table in the sidebar to get started.": "Sélectionnez une table dans la barre latérale pour commencer.",
        "selected result(s) on": "Résultats sélectionnés sur",
        "Show Filters": "Montrer les filtres",
        "Show referencing tables": "Afficher les tables de référence",
        "Showing": "Projection",
        "Sort by": "Trier par",
        "Sort": "Trier",
        "Submit": "Soumettre",
        "Tables": "les tables",
        "Thanks for using Piccolo Admin.": "Merci d'utiliser PicColo Admin.",
        "Update": "Mise à jour",
        "Use again": "Utiliser à nouveau",
        "Username": "Nom d'utilisateur",
        "Version": "Version",
        "Weeks": "Semaines",
        "Welcome to": "Bienvenue à",
        "with a matching": "avec une correspondance",
    },
)


SPANISH = Translation(
    language_name="Spanish",
    language_code="es",
    translations={
        "About": "Sobre",
        "Add Row": "Añadir fila",
        "Add": "Añadir",
        "Apply": "Aplicar",
        "Ascending": "Ascendente",
        "Back to home page": "Volver a la página de inicio",
        "Back": "atrás",
        "Change password": "Cambia la contraseña",
        "Clear filters": "Eliminar filtros",
        "Close": "Cerca",
        "Create": "Crear",
        "Current password": "Contraseña actual",
        "Dark Mode": "Modo oscuro",
        "Days": "Días",
        "Delete": "Borrar",
        "Descending": "Descendente",
        "Edit": "Editar",
        "Export CSV": "CSV de exportación",
        "Filter": "Filtrar",
        "Form submitted": "Formulario enviado",
        "Forms": "Formularios",
        "Go to page": "Ir a la página",
        "Hide Filters": "Ocultar filtros",
        "Hide referencing tables": "Ocultar tablas de referencia",
        "Home": "Hogar",
        "Hours": "Horas",
        "Light Mode": "Modo de luz",
        "Loading": "Cargando",
        "Log out": "Cerrar sesión",
        "Login": "Acceso",
        "Minutes": "Minutos",
        "New password confirmation": "Nueva confirmación de contraseña",
        "New password": "Nueva contraseña",
        "New value": "Nuevo valor",
        "No results found": "No se han encontrado resultados",
        "of": "de",
        "page": "página",
        "Password": "Clave",
        "result(s)": "resultado(s)",
        "rows": "hilera",
        "Save": "Ahorrar",
        "Seconds": "Segundos",
        "Select a column to update": "Seleccione una columna para actualizar",
        "Select a Column": "Seleccione una columna",
        "Select a table in the sidebar to get started.": "Seleccione una tabla en la barra lateral para comenzar.",
        "selected result(s) on": "Resultados seleccionados en",
        "Show Filters": "Mostrar filtros",
        "Show referencing tables": "Mostrar tablas de referencia",
        "Showing": "Demostración",
        "Sort by": "Ordenar por",
        "Sort": "Clasificar",
        "Submit": "Enviar",
        "Tables": "Tablas",
        "Thanks for using Piccolo Admin.": "Gracias por usar Piccolo Admin.",
        "Update": "Actualizar",
        "Use again": "Usar de nuevo",
        "Username": "Nombre de usuario",
        "Version": "Versión",
        "Weeks": "Semanas",
        "Welcome to": "Bienvenido a",
        "with a matching": "con un juego",
    },
)

FINNISH = Translation(
    language_name="Finnish",
    language_code="fi",
    translations={
        "About": "Tietoa",
        "Add Row": "Lisää rivi",
        "Add": "Lisää",
        "Apply": "Aseta",
        "Ascending": "Nouseva",
        "Back to home page": "Takaisin pääsivulle",
        "Back": "Takaisin",
        "Change password": "Vaihda salasana",
        "Clear filters": "Nollaa suodattimet",
        "Close": "Sulje",
        "Create": "Luo",
        "Current password": "Nykyinen salasana",
        "Dark Mode": "Tumma tila",
        "Days": "Päivät",
        "Delete": "Poista",
        "Descending": "Laskeva",
        "Edit": "Muokkaa",
        "Export CSV": "Vie CSV",
        "Filter": "Suodata",
        "Form submitted": "Lomake lähetetty",
        "Forms": "Lomakkeet",
        "Go to page": "Mene sivulle",
        "Hide Filters": "Piilota suodattimet",
        "Hide referencing tables": "Piilota viitetaulukot",
        "Home": "Koti",
        "Hours": "Tunnit",
        "Light Mode": "Vaalea tila",
        "Loading": "Latautuu",
        "Log out": "Kirjaudu ulos",
        "Login": "Kirjaudu",
        "Minutes": "Minuutit",
        "New password confirmation": "Uusi salasana hyväksyntä",
        "New password": "Uusi salasana",
        "New value": "Uusi arvo",
        "No results found": "Ei löytynyt tuloksia",
        "of": "osa",
        "page": "sivu",
        "Password": "Salasana",
        "result(s)": "tulokset",
        "rows": "rivit",
        "Save": "Tallenna",
        "Seconds": "Sekunnit",
        "Select a column to update": "Valitse päivitettävä pystyrivi",
        "Select a Column": "Valitse pystyrivi",
        "Select a table in the sidebar to get started.": "Valitse taulu sivupalkista aloittaaksesi.",
        "selected result(s) on": "valitut tulokset",
        "Show Filters": "Näytä suodattimet",
        "Show referencing tables": "Näytä viitetaulut",
        "Showing": "Näkyvissä",
        "Sort by": "Lajittele jnkn mukaan",
        "Sort": "Lajittele",
        "Submit": "Lähetä",
        "Tables": "Taulukot",
        "Thanks for using Piccolo Admin.": "Kiitos, että käytät Piccolo Adminia.",
        "Update": "Päivitä",
        "Use again": "Käytä uudelleen",
        "Username": "Käyttäjänimi",
        "Version": "Versio",
        "Weeks": "Viikot",
        "Welcome to": "Tervetuloa",
        "with a matching": "osumalla",
    },
)

RUSSIAN = Translation(
    language_name="Russian",
    language_code="ru",
    translations={
        "About": "О продукте",
        "Add Row": "Добавить строку",
        "Add": "Добавить",
        "Apply": "Подтвердить",
        "Ascending": "По возрастанию",
        "Back to home page": "Вернуться на главную страницу",
        "Back": "Назад",
        "Change password": "Сменить пароль",
        "Clear filters": "Сбросить фильтры",
        "Close": "Закрыть",
        "Create": "Создать",
        "Current password": "Текущий пароль",
        "Dark Mode": "Темная тема",
        "Days": "Дни",
        "Delete": "Удалить",
        "Descending": "По убыванию",
        "Edit": "Изменить",
        "Export CSV": "Экспорт в CSV",
        "Filter": "Фильтр",
        "Form submitted": "Форма отправлена",
        "Forms": "Формы",
        "Go to page": "Перейти к странице",
        "Hide Filters": "Скрыть фильтры",
        "Hide referencing tables": "Скрыть связаные таблицы",
        "Home": "Главная",
        "Hours": "Часы",
        "Light Mode": "Светлая тема",
        "Loading": "Загрузка",
        "Log out": "Выйти",
        "Login": "Войти",
        "Minutes": "Минуты",
        "New password confirmation": "Повторите новый пароль",
        "New password": "Новый пароль",
        "New value": "Новое значение",
        "No results found": "Ничего не найдено",
        "of": "из",
        "page": "страницу(е)",
        "Password": "Пароль",
        "result(s)": "строк(и)",
        "rows": "строки",
        "Save": "Сохранить",
        "Seconds": "Секунды",
        "Select a column to update": "Выберите столбец для обновления",
        "Select a Column": "Выберите столбец",
        "Select a table in the sidebar to get started.": "Выберите таблицу в боковой панели.",
        "selected result(s) on": "выбрано на",
        "Show Filters": "Показать фильтры",
        "Show referencing tables": "Показать связанные таблицы",
        "Showing": "Показано",
        "Sort by": "Сортировать по",
        "Sort": "Сортировка",
        "Submit": "Отправить",
        "Tables": "Таблицы",
        "Thanks for using Piccolo Admin.": "Спасибо, что используете Piccolo Admin.",
        "Update": "Обновить",
        "Use again": "Использовать снова",
        "Username": "Логин",
        "Version": "Версия",
        "Weeks": "Недели",
        "Welcome to": "Добро пожаловать в",
        "with a matching": "c соответствующим",
    },
)

UKRAINIAN = Translation(
    language_name="Ukrainian",
    language_code="uk",
    translations={
        "About": "Про продукт",
        "Add Row": "Додати рядок",
        "Add": "Додати",
        "Apply": "Застосувати",
        "Ascending": "За зростанням",
        "Back to home page": "Повернутися на головну сторінку",
        "Back": "Назад",
        "Change password": "Змінити пароль",
        "Clear filters": "Очистити фільтри",
        "Close": "Закрити",
        "Create": "Створити",
        "Current password": "Поточний пароль",
        "Dark Mode": "Темна тема",
        "Days": "Дні",
        "Delete": "Видалити",
        "Descending": "За спаданням",
        "Edit": "Редагувати",
        "Export CSV": "Експорт в CSV",
        "Filter": "Фільтр",
        "Form submitted": "Форму відправлено",
        "Forms": "Форми",
        "Go to page": "Перейти на сторінку",
        "Hide Filters": "Приховати фільтри",
        "Hide referencing tables": "Приховати пов'язані таблиці",
        "Home": "Головна",
        "Hours": "Години",
        "Light Mode": "Світла тема",
        "Loading": "Завантаження",
        "Log out": "Вийти",
        "Login": "Увійти",
        "Minutes": "Хвилини",
        "New password confirmation": "Підтвердіть новий пароль",
        "New password": "Новий пароль",
        "New value": "Нове значення",
        "No results found": "Нічого не знайдено",
        "of": "з",
        "page": "сторінки",
        "Password": "Пароль",
        "result(s)": "результат(ів)",
        "rows": "рядки",
        "Save": "Зберегти",
        "Seconds": "Секунди",
        "Select a column to update": "Виберіть стовпчик для оновлення",
        "Select a Column": "Виберіть стовпчик",
        "Select a table in the sidebar to get started.": "Виберіть таблицю в бічній панелі.",
        "selected result(s) on": "вибрано на",
        "Show Filters": "Показати фільтри",
        "Show referencing tables": "Показати пов'язані таблиці",
        "Showing": "Показано",
        "Sort by": "Сортувати за",
        "Sort": "Сортування",
        "Submit": "Відправити",
        "Tables": "Таблиці",
        "Thanks for using Piccolo Admin.": "Дякуємо, що використовуєте Piccolo Admin. Рускій корабль іде нах@й!",
        "Update": "Оновити",
        "Use again": "Використати ще раз",
        "Username": "Ім'я користувача",
        "Version": "Версія",
        "Weeks": "Тижні",
        "Welcome to": "Ласкаво просимо до",
        "with a matching": "з пов'язаним",
    },
)

SIMPLIFIED_CHINESE = Translation(
    language_name="Simplified Chinese",
    language_code="zh-CN",
    translations={
        "About": "关于",
        "Add Row": "添加行",
        "Add": "添加",
        "Apply": "应用",
        "Ascending": "正序",
        "Back to home page": "返回主页",
        "Back": "返回",
        "Change password": "修改密码",
        "Clear filters": "清除过滤器",
        "Close": "关闭",
        "Create": "创建",
        "Current password": "当前密码",
        "Dark Mode": "夜间模式",
        "Days": "天",
        "Delete": "删除",
        "Descending": "倒序",
        "Edit": "编辑",
        "Export CSV": "导出 CSV",
        "Filter": "过滤器",
        "Form submitted": "已提交的表单",
        "Forms": "表单",
        "Go to page": "转到页面",
        "Hide Filters": "隐藏过滤器",
        "Hide referencing tables": "隐藏引用的表",
        "Home": "主页",
        "Hours": "小时",
        "Light Mode": "白天模式",
        "Loading": "加载中",
        "Log out": "登出",
        "Login": "登录",
        "Minutes": "分钟",
        "New password confirmation": "确认新密码",
        "New password": "新密码",
        "New value": "新的值",
        "No results found": "没有找到结果",
        "of": "于",  # prep. is hard to translate in Chinese
        "page": "页",
        "Password": "密码",
        "result(s)": "结果",
        "rows": "行",
        "Save": "保存",
        "Seconds": "秒",
        "Select a column to update": "选择要更新的列",
        "Select a Column": "选择一个列",
        "Select a table in the sidebar to get started.": "请在侧栏选择一个表来开始编辑",
        "selected result(s) on": "选择的结果",
        "Show Filters": "显示过滤器",
        "Show referencing tables": "显示引用的表",
        "Showing": "正在显示",
        "Sort by": "使用排序",  # it should be 使用...排序
        "Sort": "排序",
        "Submit": "提交",
        "Tables": "表",
        "Thanks for using Piccolo Admin.": "感谢您使用 Piccolo 管理界面",
        "Update": "更新",
        "Use again": "再次使用",
        "Username": "用户名",
        "Version": "版本",
        "Weeks": "周",
        "Welcome to": "欢迎来到",
        "with a matching": "匹配了",  # it should be 与...匹配
    },
)

TRANSLATIONS: t.List[Translation] = [
    ENGLISH,
    CROATIAN,
    FINNISH,
    FRENCH,
    GERMAN,
    PORTUGUESE,
    SIMPLIFIED_CHINESE,
    SPANISH,
    RUSSIAN,
    UKRAINIAN,
    WELSH,
]



"""
An example of how to configure and run the admin.

Can be run from the command line using `python -m piccolo_admin.example`,
or `admin_demo`.
"""
import asyncio
import datetime
import decimal
import enum
import os
import random
import smtplib
import typing as t

import targ
from hypercorn.asyncio import serve
from hypercorn.config import Config
from piccolo.apps.user.tables import BaseUser
from piccolo.columns.column_types import (
    JSON,
    UUID,
    Array,
    BigInt,
    Boolean,
    Date,
    ForeignKey,
    Integer,
    Interval,
    Numeric,
    Real,
    SmallInt,
    Text,
    Time,
    Timestamp,
    Varchar,
)
from piccolo.columns.readable import Readable
from piccolo.engine.postgres import PostgresEngine
from piccolo.engine.sqlite import SQLiteEngine
from piccolo.table import Table, create_db_tables_sync, drop_db_tables_sync
from piccolo_api.media.local import LocalMediaStorage
from piccolo_api.media.s3 import S3MediaStorage
from piccolo_api.session_auth.tables import SessionsBase
from pydantic import BaseModel, validator

from piccolo_admin.endpoints import FormConfig, TableConfig, create_admin
from piccolo_admin.example_data import (
    DIRECTORS,
    MOVIE_WORDS,
    MOVIES,
    STUDIOS,
    TICKETS,
)

try:
    """
    If you want to try out S3, create a .env file in this folder, with the
    following contents (inserting your S3 credentials where appropriate):

        AWS_ACCESS_KEY_ID=abc123
        AWS_SECRET_ACCESS_KEY=abc123
        BUCKET_NAME=bucket123

    These values can also be added if required:

        ENDPOINT_URL=s3.cloudprovider.com
        REGION_NAME=my-region

    The ``Director.photo`` column will then use S3 for storage.

    """

    import dotenv

    dotenv.load_dotenv(os.path.join(os.path.dirname(__file__), ".env"))

    AWS_ACCESS_KEY_ID = os.environ.get("AWS_ACCESS_KEY_ID")
    AWS_SECRET_ACCESS_KEY = os.environ.get("AWS_SECRET_ACCESS_KEY")
    BUCKET_NAME = os.environ.get("BUCKET_NAME")

    USE_S3 = all(
        (
            AWS_SECRET_ACCESS_KEY,
            AWS_SECRET_ACCESS_KEY,
            BUCKET_NAME,
        )
    )

    if USE_S3:
        print("Using S3")

        S3_CONFIG = {
            "aws_access_key_id": AWS_ACCESS_KEY_ID,
            "aws_secret_access_key": AWS_SECRET_ACCESS_KEY,
        }

        ENDPOINT_URL = os.environ.get("ENDPOINT_URL")
        if ENDPOINT_URL:
            S3_CONFIG["endpoint_url"] = ENDPOINT_URL

        REGION_NAME = os.environ.get("REGION_NAME")
        if REGION_NAME:
            S3_CONFIG["region_name"] = REGION_NAME

except ImportError:
    USE_S3 = False


MEDIA_ROOT = os.path.join(os.path.dirname(__file__), "example_media")
if not USE_S3 and not os.path.exists(MEDIA_ROOT):
    os.mkdir(MEDIA_ROOT)


class Sessions(SessionsBase):
    pass


class User(BaseUser, tablename="piccolo_user"):
    pass


class Director(Table, help_text="The main director for a movie."):
    class Gender(str, enum.Enum):
        male = "m"
        female = "f"
        non_binary = "n"

    class Skill(str, enum.Enum):
        producer = "producer"
        actor = "actor"
        director_of_photography = "director of photography"
        special_effects = "special effects"

    name = Varchar(length=300, null=False)
    years_nominated = Array(
        base_column=Integer(),
        help_text=(
            "Which years this director was nominated for a best director "
            "Oscar."
        ),
    )
    gender = Varchar(length=1, choices=Gender)
    photo = Varchar()
    additional_skills = Array(
        base_column=Varchar(),
        choices=Skill,
        help_text=(
            "Additional skills which the director has besides directing."
        ),
    )

    @classmethod
    def get_readable(cls):
        return Readable(template="%s", columns=[cls.name])


class Studio(Table, help_text="A movie studio."):
    pk = UUID(primary_key=True)
    name = Varchar(unique=True)
    facilities = JSON()

    @classmethod
    def get_readable(cls):
        return Readable(template="%s", columns=[cls.name])


class Movie(Table):
    class Genre(int, enum.Enum):
        fantasy = 1
        sci_fi = 2
        documentary = 3
        horror = 4
        action = 5
        comedy = 6
        romance = 7
        musical = 8

    name = Varchar(length=300)
    rating = Real(help_text="The rating on IMDB.")
    duration = Interval()
    director = ForeignKey(references=Director)
    oscar_nominations = Integer()
    won_oscar = Boolean()
    description = Text()
    poster = Varchar()
    screenshots = Array(base_column=Varchar())
    release_date = Date(null=True)
    box_office = Numeric(digits=(5, 1), help_text="In millions of US dollars.")
    tags = Array(base_column=Varchar())
    barcode = BigInt(default=0)
    genre = SmallInt(choices=Genre, null=True)
    studio = ForeignKey(Studio)

    @classmethod
    def get_readable(cls):
        return Readable(template="%s", columns=[cls.name])


class Ticket(Table):
    booked_by = Varchar(length=255)
    movie = ForeignKey(Movie)
    start_date = Date()
    start_time = Time()
    booked_on = Timestamp()


class BusinessEmailModel(BaseModel):
    email: str
    title: str = "Enquiry"
    content: str

    @validator("email")
    def validate_email(cls, v):
        if "@" not in v:
            raise ValueError("not valid email")
        return v


class BookingModel(BaseModel):
    email: str
    name: str
    notes: str = "N/A"

    @validator("email")
    def validate_email(cls, v):
        if "@" not in v:
            raise ValueError("not valid email")
        return v


def business_email_endpoint(request, data):
    sender = "info@example.com"
    receivers = [data.email]

    message = f"""From: Piccolo Admin <info@example.com>
    To: Colleague <{data.email}>
    Subject: {data.title}
    {data.content}
    """

    try:
        smtpObj = smtplib.SMTP("localhost:1025")
        smtpObj.sendmail(sender, receivers, message)
        print("Successfully sent email")
    except (smtplib.SMTPException, ConnectionRefusedError):
        print("Error: unable to send email")

    return "Email sent"


async def booking_endpoint(request, data):
    """
    Testing that async functions works.
    """
    sender = "info@example.com"
    receivers = [data.email]

    message = f"""From: Bookings <info@example.com>
    To: To Friend <{data.email}>
    Subject: {data.name} booking
    {data.notes}
    """

    try:
        smtpObj = smtplib.SMTP("localhost:1025")
        smtpObj.sendmail(sender, receivers, message)
        print("Successfully sent email")
    except (smtplib.SMTPException, ConnectionRefusedError):
        print("Error: unable to send email")

    return "Booking complete"


TABLE_CLASSES: t.Tuple[t.Type[Table], ...] = (
    Director,
    Movie,
    Studio,
    User,
    Sessions,
    Ticket,
)


movie_config = TableConfig(
    table_class=Movie,
    visible_columns=[
        Movie._meta.primary_key,
        Movie.name,
        Movie.rating,
        Movie.director,
        Movie.poster,
        Movie.tags,
        Movie.screenshots,
    ],
    visible_filters=[
        Movie.name,
        Movie.rating,
        Movie.director,
        Movie.duration,
        Movie.genre,
        Movie.screenshots,
        Movie.poster,
        Movie.release_date,
    ],
    rich_text_columns=[Movie.description],
    media_storage=(
        LocalMediaStorage(
            column=Movie.poster,
            media_path=os.path.join(MEDIA_ROOT, "movie_poster"),
        ),
        LocalMediaStorage(
            column=Movie.screenshots,
            media_path=os.path.join(MEDIA_ROOT, "movie_screenshots"),
        ),
    ),
)

director_config = TableConfig(
    table_class=Director,
    visible_columns=[
        Director._meta.primary_key,
        Director.name,
        Director.gender,
        Director.photo,
    ],
    media_storage=(
        S3MediaStorage(
            column=Director.photo,
            bucket_name=t.cast(str, BUCKET_NAME),
            folder_name="director_photo",
            connection_kwargs=S3_CONFIG,
        )
        if USE_S3
        else LocalMediaStorage(
            column=Director.photo,
            media_path=os.path.join(MEDIA_ROOT, "photo"),
        ),
    ),
)

APP = create_admin(
    [movie_config, director_config, Studio, Ticket],
    forms=[
        FormConfig(
            name="Business email form",
            pydantic_model=BusinessEmailModel,
            endpoint=business_email_endpoint,
            description="Send an email to a business associate.",
        ),
        FormConfig(
            name="Booking form",
            pydantic_model=BookingModel,
            endpoint=booking_endpoint,
            description="Make a booking for a customer.",
        ),
    ],
    auth_table=User,
    session_table=Sessions,
)


def set_engine(engine: str = "sqlite"):
    if engine == "postgres":
        db = PostgresEngine(config={"database": "piccolo_admin"})
    else:
        sqlite_path = os.path.join(os.path.dirname(__file__), "example.sqlite")
        db = SQLiteEngine(path=sqlite_path)  # type: ignore

    for table_class in TABLE_CLASSES:
        table_class._meta._db = db


def create_schema(persist: bool = False):
    if not persist:
        drop_db_tables_sync(*TABLE_CLASSES)

    create_db_tables_sync(*TABLE_CLASSES, if_not_exists=True)


def populate_data(inflate: int = 0, engine: str = "sqlite"):
    """
    Populate the database with some example data.

    :param inflate:
        If set, this number of extra rows are inserted containing dummy data.
        This is useful for testing.

    """
    # Add some rows
    Director.insert(*[Director(**d) for d in DIRECTORS]).run_sync()
    Movie.insert(*[Movie(**m) for m in MOVIES]).run_sync()
    Studio.insert(*[Studio(**s) for s in STUDIOS]).run_sync()
    Ticket.insert(*[Ticket(**t) for t in TICKETS]).run_sync()

    if engine == "postgres":
        # We need to update the sequence, as we explicitly set the IDs for the
        # directors we just inserted
        Director.raw(
            "SELECT setval('director_id_seq', max(id)) FROM director"
        ).run_sync()

    # Create a user for testing login
    user = User(
        username="piccolo",
        password="piccolo123",
        email="admin@test.com",
        admin=True,
        active=True,
        superuser=True,
    )
    user.save().run_sync()

    new_user = User(
        username="john",
        password="john123",
        email="john@test.com",
        admin=True,
        active=True,
        superuser=False,
    )
    new_user.save().run_sync()

    if inflate:
        try:
            import faker
        except ImportError:
            print(
                "Install faker to use this feature: "
                "`pip install piccolo_admin[faker]`"
            )
        else:
            fake = faker.Faker()
            remaining = inflate
            chunk_size = 100

            while remaining > 0:
                if remaining < chunk_size:
                    chunk_size = remaining
                    remaining = 0
                else:
                    remaining = remaining - chunk_size

                directors = []
                genders = ["m", "f", "n"]
                for _ in range(chunk_size):
                    gender = random.choice(genders)
                    if gender == "m":
                        name = fake.name_male()
                    elif gender == "f":
                        name = fake.name_female()
                    else:
                        name = fake.name_nonbinary()
                    directors.append(Director(name=name, gender=gender))

                Director.insert(*directors).run_sync()

                director_ids = (
                    Director.select(Director._meta.primary_key)
                    .order_by(Director._meta.primary_key, ascending=False)
                    .limit(chunk_size)
                    .output(as_list=True)
                    .run_sync()
                )

                movies = []
                genres = [i.value for i in Movie.Genre]
                for _ in range(chunk_size):
                    oscar_nominations = random.sample(
                        [0, 0, 0, 0, 0, 1, 1, 3, 5], 1
                    )[0]
                    won_oscar = oscar_nominations > 0
                    rating = (
                        random.randint(80, 100)
                        if won_oscar
                        else random.randint(1, 100)
                    ) / 10

                    movie = Movie(
                        name="{} {}".format(
                            fake.word().title(),
                            fake.word(ext_word_list=MOVIE_WORDS),
                        ),
                        rating=rating,
                        duration=datetime.timedelta(
                            minutes=random.randint(60, 210)
                        ),
                        director=random.sample(director_ids, 1)[0],
                        oscar_nominations=oscar_nominations,
                        won_oscar=won_oscar,
                        description=fake.sentence(30),
                        release_date=fake.date(),
                        box_office=decimal.Decimal(
                            str(random.randint(10, 1500) / 10)
                        ),
                        barcode=random.randint(1_000_000_000, 9_999_999_999),
                        genre=random.choice(genres),
                    )
                    movies.append(movie)

                Movie.insert(*movies).run_sync()


def run(persist: bool = False, engine: str = "sqlite", inflate: int = 0):
    """
    Start the Piccolo admin.

    :param persist:
        If True, we don't rebuild all of the data each time.
    :param engine:
        Options are sqlite and postgres. By default sqlite is used.
    :param inflate:
        If set, this number of extra rows are inserted containing dummy data.
        This is useful when you need to test with lots of data. Example
        `--inflate=10000`.

    """
    set_engine(engine)
    create_schema(persist=persist)

    if not persist:
        populate_data(inflate=inflate, engine=engine)

    # Server
    class CustomConfig(Config):
        use_reloader = True
        accesslog = "-"

    asyncio.run(serve(APP, CustomConfig()))


def main():
    cli = targ.CLI(description="Piccolo Admin")
    cli.register(run)
    cli.run(solo=True)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import itertools
import os
import typing as t

from setuptools import find_packages, setup

DIRECTORY = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(DIRECTORY, "requirements/requirements.txt")) as f:
    contents = f.read()
    REQUIREMENTS = [i.strip() for i in contents.strip().split("\n")]


with open(os.path.join(DIRECTORY, "README.md")) as f:
    LONG_DESCRIPTION = f.read()


EXTRAS = ["faker", "s3"]


def parse_requirement(req_path: str) -> t.List[str]:
    """
    Parses a requirement file - returning a list of contents.

    Example::

        parse_requirement('requirements.txt')       # requirements/requirements.txt
        parse_requirement('extras/playground.txt')  # requirements/extras/playground.txt

    :returns: A list of requirements specified in the file.

    """  # noqa: E501
    with open(os.path.join(DIRECTORY, "requirements", req_path)) as f:
        contents = f.read()
        return [i.strip() for i in contents.strip().split("\n")]


def extras_require() -> t.Dict[str, t.List[str]]:
    """
    Parse requirements in requirements/extras directory
    """
    extra_requirements = {
        extra: parse_requirement(os.path.join("extras", f"{extra}.txt"))
        for extra in EXTRAS
    }

    extra_requirements["all"] = list(
        itertools.chain.from_iterable(extra_requirements.values())
    )

    return extra_requirements


def get_version():
    with open(os.path.join(DIRECTORY, "piccolo_admin", "version.txt")) as f:
        return f.read().strip()


setup(
    name="piccolo_admin",
    version=get_version(),
    description="A powerful and modern admin interface / CMS, powered by Piccolo and ASGI.",  # noqa: E501
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Daniel Townsend",
    author_email="dan@dantownsend.co.uk",
    python_requires=">=3.7.0",
    url="https://github.com/piccolo-orm/piccolo_admin",
    packages=find_packages(exclude=("tests",)),
    install_requires=REQUIREMENTS,
    extras_require=extras_require(),
    license="MIT",
    include_package_data=True,
    entry_points={
        "console_scripts": ["admin_demo = piccolo_admin.example:main"],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Database :: Front-Ends",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Content Management System",  # noqa: E501
    ],
)
